#!/usr/bin/env python3
from dactyl.common import *
import argparse

class DactylCLIParser:
    UTIL_BUILD = "Generate static site from markdown and templates."
    UTIL_LINKS = "Check files in this repository for broken links."
    UTIL_STYLE = "Check content files for style issues."

    def __init__(self, utility):
        """Specify commandline usage and parse arguments"""
        parser = argparse.ArgumentParser(description=utility)

        noisiness = parser.add_mutually_exclusive_group(required=False)
        noisiness.add_argument("--quiet", "-q", action="store_true",
                            help="Suppress status messages")
        noisiness.add_argument("--debug", action="store_true",
                            help="Print debug-level log messages")

        parser.add_argument("--config", "-c", type=str,
                            help="Specify path to an alternate config file.")
        parser.add_argument("--version", "-v", action="store_true",
                            help="Print version information and exit.")
        parser.add_argument("--bypass_errors", "-b", action="store_true",
                            help="Continue if recoverable errors occur")

        if utility in (self.UTIL_BUILD, self.UTIL_STYLE):
            parser.add_argument("--target", "-t", type=str,
                help="Use the specified target (from the config file).")

        if utility == self.UTIL_BUILD:
            build_mode = parser.add_mutually_exclusive_group(required=False)
            build_mode.add_argument("--pdf", nargs="?", type=str,
                                const=PDF_USE_DEFAULT, default=NO_PDF,
                                help="Output a PDF to this file. Requires Prince.")
            build_mode.add_argument("--md", action="store_true",
                                help="Output markdown only")
            build_mode.add_argument("--html", action="store_true", default=True,
                                help="Output HTML files (the default)")
            build_mode.add_argument("--es", action="store_true",
                                help="Output JSON for ElasticSearch upload")
            # HTML is the default mode

            static_files = parser.add_mutually_exclusive_group(required=False)
            static_files.add_argument("--copy_static", "-s", action="store_true",
                        help="Copy all static files to the out dir",
                        default=False)
            static_files.add_argument("--no_static", "-S", action="store_true",
                        help="Don't copy any static files to the out dir",
                        default=False)
            static_files.add_argument("--template_static", "-T", action="store_true",
                        help="Copy only templates' static files to the out dir",
                        default=False)
            static_files.add_argument("--content_static", "-C", action="store_true",
                        help="Copy only the content's static files to the out dir",
                        default=False)
            parser.add_argument("--es_upload", nargs="?", type=str,
                                const=ES_USE_DEFAULT, default=NO_ES_UP,
                                help="Upload documents to ElasticSearch cluster "+
                                "at this URL (http://localhost:9200 by default). "+
                                "Ignored when making PDFs.")
            parser.add_argument("--leave_temp_files", action="store_true",
                                help="Leave temp files in place (for debugging or "+
                                "manual PDF generation). Ignored when using --watch",
                                default=False)
            parser.add_argument("--list_targets_only", "-l", action="store_true",
                                help="Don't build anything, just display list of "+
                                "known targets from the config file.")
            parser.add_argument("--only", type=str, help=".md or .html filename of a "+
                                "single page in the config to build alone.")
            parser.add_argument("--out_dir", "-o", type=str,
                            help="Output to this folder (overrides config file)")
            parser.add_argument("--pages", type=str, help="Markdown file(s) to build "+\
                                "that aren't described in the config.", nargs="+")
            parser.add_argument("--openapi", type=str, help="OpenAPI spec file "+
                                "to generate docs from.")
            parser.add_argument("--no_cover", "-n", action="store_true",
                                help="Don't automatically add a cover / index file.")
            parser.add_argument("--skip_preprocessor", action="store_true", default=False,
                                help="Don't pre-process Jinja syntax in markdown files")
            parser.add_argument("--template_strict_undefined", action="store_true",
                                help="Raise an error on undefined variables in "+
                                "template syntax.")
            parser.add_argument("--pp_strict_undefined", action="store_true",
                                help="Raise an error on undefined variables in "+
                                "preprocessor syntax.")
            parser.add_argument("--title", type=str, help="Override target display "+\
                                "name. Useful when passing multiple args to --pages.")
            parser.add_argument("--vars", type=str, help="A YAML or JSON file with vars "+
                                "to add to the target so the preprocessor and "+
                                "templates can reference them.")
            parser.add_argument("--watch", "-w", action="store_true", default=False,
                                help="Watch for changes and re-generate output. "+\
                                "This runs until force-quit.")
            parser.add_argument("--http_port", type=int, default=DEFAULT_SERVER_PORT,
                                help="Use this port for HTTP server (when "+\
                                "building PDFs.) Use '0' for no server (may not "+\
                                "work well with PDFs containing absolute links)")

        elif utility == self.UTIL_LINKS:
            parser.add_argument("-o", "--offline", action="store_true",
                help="Check local anchors only")
            parser.add_argument("-d", "--dir", type=str, default=None,
                help="Check the specified dir (otherwise use configured out_dir)")
            parser.add_argument("-s", "--strict", action="store_true",
                help="Exit with error even on known problems")
            parser.add_argument("-n", "--no_final_retry", action="store_true",
                help="Don't wait and retry failed remote links at the end.")
            prefixes = parser.add_mutually_exclusive_group(required=False)
            prefixes.add_argument("-p", "--prefix", type=str, default="/",
                help="Assume site starts at this path. Must start with '/'")
            prefixes.add_argument("--no_prefix", action="store_true", default=False,
                help="Don't check absolute links")


        self.cli_args = parser.parse_args()



#!/usr/bin/env python3
from dactyl.common import *

import requests
from bs4 import BeautifulSoup
from time import time, sleep

from dactyl.config import DactylConfig
from dactyl.cli import DactylCLIParser

TIMEOUT_SECS = 9.1
CHECK_IN_INTERVAL = 30
FINAL_RETRY_DELAY = 4 * CHECK_IN_INTERVAL

TYPE_LINK = "link"
TYPE_IMAGE = "image"

soupsCache = {}
def getSoup(in_file):
    if in_file in soupsCache.keys():
        soup = soupsCache[in_file]
    else:
        with open(in_file, 'r', encoding="utf-8") as f:
            soup = BeautifulSoup(f.read(), "html.parser")
            soupsCache[in_file] = soup
    return soup


def check_for_unparsed_reference_links(soup):
    #unmatched_reflink_regex = re.compile(r"\[[^\]]+\]\[(\w| )*\]")
    unmatched_reflink_regex = re.compile(r"(\[[^\]]+)?\]\[(\w| )*\]")
    unparsed_links = []
    for s in soup.strings:
        m = re.search(unmatched_reflink_regex, s)
        if m:
            unparsed_links.append(m.group(0))
    return unparsed_links


remote_url_cache = {
    #str url: bool was_good
}
def check_remote_url(endpoint, in_file, hreftype=TYPE_LINK):
    if endpoint in remote_url_cache.keys():
        if remote_url_cache[endpoint]:
            logger.debug("Skipping cached %s %s" % (hreftype, endpoint))
            return (1, True)
        else:
            # We already confirmed this was broken, so just add another instance
            logger.warning("Broken %s '%s' appears again in %s" %
                    (hreftype, endpoint, in_file))
            return (1, False)

    logger.info("Testing remote %s '%s'" % (hreftype, endpoint))
    try:
        code = requests.head(endpoint, timeout=TIMEOUT_SECS).status_code
    except Exception as e:
        logger.warning("Error occurred: %s" % repr(e))
        code = 500

    if code == 405 or code == 404:
        #HEAD didn't work, maybe GET will?
        try:
            code = requests.get(endpoint, timeout=TIMEOUT_SECS).status_code
        except Exception as e:
          logger.warning("... Error occurred: %s" % repr(e))
          code = 500

    if code < 200 or code >= 400:
        logger.warning("... Broken remote %s in %s to '%s'" %
                (hreftype, in_file, endpoint))
        remote_url_cache[endpoint] = False
        return (1, False)
    else:
        logger.info("... success.")
        remote_url_cache[endpoint] = True
        return (1, True)


def check_href(endpoint, in_file, dirpath, top_dir, offline, hreftype=TYPE_LINK, site_prefix=None):
    """
    Check a given href to see if it's a working link (coming from )
    Use hreftype=TYPE_IMAGE for modified behavior/logging for images.
    Returns (int was_checked, bool was_good) where was_checked indicates how
        many links were actually checked (1 or 0) and was_good indicates
        whether the link was found if checked.
        If the link was not checked, was_good is False.
    """

    if not endpoint.strip():
        logger.warning("Empty %s in %s" % (hreftype,in_file))
        return (1, False)

    if endpoint in config["known_broken_links"]:
        logger.warning("Skipping known broken %s '%s' in %s" %
                (hreftype, endpoint, in_file))
        return (0, False)

    if endpoint == "#":
        if hreftype == TYPE_IMAGE:
            logger.warning("Invalid image URL: '#' in %s" % in_file)
            return (1, False)
        logger.debug("Skipping empty anchor link in %s" % in_file)
        return (0, False)

    if "mailto:" in endpoint:
        if hreftype == TYPE_IMAGE:
            logger.warning("Invalid image URL: '%s' in %s" % (endpoint, in_file))
            return (1, False)
        logger.warning("Skipping email link in %s to %s" %
                (in_file, endpoint))
        return (0, False)

    if "://" in endpoint or endpoint[:2] == "//":
        # Remote link.
        if offline:
            logger.info("Offline - Skipping remote URL %s" % (endpoint))
            return (0, False)

        return check_remote_url(endpoint, in_file, hreftype)
    else:
        # Local Anchor link; check presence of target file & anchor too.
        return check_local_file(endpoint, in_file, dirpath, top_dir, site_prefix)


def check_local_file(endpoint, in_file, dirpath, top_dir, site_prefix):
    # First, split the link up into parts
    if "#" in endpoint:
        filename, anchor = endpoint.split("#",1)
    else:
        filename, anchor = endpoint,""
    if "?" in filename:
        filename, query = filename.split("?", 1)
    else:
        filename, query = filename, ""

    if filename == "":
        # Anchor only; linked file is the current one.
        full_file_path = in_file
    elif filename[0] == "/":
        # Absolute path.
        if site_prefix and filename.startswith(site_prefix):
            logger.debug("Trimming site prefix (%s) from absolute link..." %
                    site_prefix)
            full_file_path = os.path.join(top_dir, filename[len(site_prefix):])
        else:
            # Can't properly test absolute links without knowing where the
            #   server root will be, so skip this
            logger.warning("Skipping absolute link in %s to '%s'" %
                    (in_file, endpoint))
            return (0, False)
    else:
        # Relative path.
        full_file_path = os.path.join(dirpath, filename)

    # Assume index.html if path is a directory.
    # This won't work for certain unusual server configurations or if you're
    # building PHP files or something weird like that
    if os.path.isdir(full_file_path):
        full_file_path = os.path.join(full_file_path, "index.html")

    logger.info("Testing local link in %s to '%s'" %
            (in_file, endpoint))

    # See if the file is even there...
    if not os.path.exists(full_file_path):
        logger.warning("... Broken local link in %s to '%s' (file not found)" %
                (in_file, endpoint))
        return (1, False)

    if not anchor:
        # No anchor to check? Then we're done.
        logger.info("... success.")
        return (1, True)

    # Skip the anchor if this is a configured "ignore anchors" file.
    # This is mostly useful if the linked page has its anchors dynamically
    # populated by JavaScript or something.
    if filename in config["ignore_anchors_in"]:
        logger.warning("Ignoring anchor '%s' in page %s" %
                (endpoint,filename))
        return (0, False)

    targetSoup = getSoup(full_file_path)
    if not targetSoup.find(id=anchor) and not targetSoup.find(
                "a",attrs={"name":anchor}):
        logger.warning("... Broken anchor link in %s to %s" %
                (in_file, endpoint))
        return (1, False)
    else:
        logger.info("... anchor found.")
        return (1, True)


def check_directory(top_dir, offline=False, site_prefix=None):
    """
    Walk the given out_path for .html files and check links in each of them.
    """
    externalCache = []
    broken_links = []
    num_links_checked = 0
    last_checkin = time()
    for dirpath, dirnames, filenames in os.walk(top_dir):
        if time() - last_checkin > CHECK_IN_INTERVAL:
            ## Print output periodically so Jenkins/etc. don't kill the job
            last_checkin = time()
            print("... still working (dirpath: %s) ..." % dirpath)
        if "template_path" in config and \
               os.path.abspath(dirpath) == os.path.abspath(config["template_path"]):
            # don't try to parse and linkcheck the templates
            logger.warning("Skipping link checking for template path %s" % dirpath)
            continue
        for fname in filenames:
            if time() - last_checkin > CHECK_IN_INTERVAL:
                last_checkin = time()
                print("... still working (file: %s) ..." % fname)

            in_file = os.path.join(dirpath, fname)
            if "/node_modules/" in in_file or ".git" in in_file:
                logger.debug("skipping ignored dir: %s" % in_file)
                continue
            if in_file.endswith(".html"):
                soup = getSoup(in_file)
                unparsed_links = check_for_unparsed_reference_links(soup)
                if unparsed_links:
                    logger.warning("Found %d unparsed Markdown reference links: %s" %
                            (len(unparsed_links), "\n... ".join(unparsed_links)))
                    [broken_links.append( (in_file, u) ) for u in unparsed_links]

                # Check <a> tags
                links = soup.find_all('a')
                for link in links:
                    if time() - last_checkin > CHECK_IN_INTERVAL:
                        last_checkin = time()
                        print("... still working (link: %s) ..." % link)
                    if "href" not in link.attrs:
                        #probably an <a name> type anchor, skip
                        continue

                    was_checked, was_good = check_href(link['href'],
                            in_file, dirpath, top_dir, offline,
                            site_prefix=site_prefix)
                    num_links_checked += was_checked
                    if was_checked and not was_good:
                        broken_links.append( (in_file, link['href']) )

                # Check <img> tags
                imgs = soup.find_all('img')
                for img in imgs:
                    if "src" not in img.attrs or not img["src"].strip():
                        logger.warning("Broken image with no src in %s" % in_file)
                        broken_links.append( (in_file, img["src"]) )
                        num_links_checked += 1
                        continue

                    was_checked, was_good = check_href(img["src"],
                            in_file, dirpath, top_dir, offline,
                            site_prefix=site_prefix, hreftype=TYPE_IMAGE)
                    num_links_checked += was_checked
                    if was_checked and not was_good:
                        broken_links.append( (in_file, img["src"]) )

    return broken_links, num_links_checked


def final_retry_links(broken_links):
    """Give the broken remote links a little while to recover in case they're just flaps"""
    broken_remote_links = [ (page,link) for page,link in broken_links
                           if re.match(r"^https?://", link) ]
    if not broken_remote_links:
        logger.info("(no http/https broken links to retry)")
        return

    print("Waiting %d seconds to retry %d broken remote links..."
                % (FINAL_RETRY_DELAY, len(broken_remote_links)))
    start_wait = time()
    elapsed = 0
    while elapsed < FINAL_RETRY_DELAY:
        sleep(CHECK_IN_INTERVAL)
        print("...")
        elapsed = time() - start_wait

    global remote_url_cache
    remote_url_cache = {}
    retry_broken = []
    for page, link in broken_remote_links:
        was_checked, was_good = check_remote_url(link, page)
        if was_good:
            logger.info("Link '%s' in page %s is back online" % (link, page))
            broken_links.remove( (page,link) )
        else:
            logger.info("Link '%s' in page %s is still down." % (link, page))



def main(cli_args):
    if cli_args.dir:
        out_path = cli_args.dir
        logger.debug("Chose top_dir '%s'" % out_path)
    else:
        out_path = config["out_path"]

    if cli_args.prefix:
        if cli_args.prefix[0] != "/":
            logger.error("Invalid --prefix value '%s'. Prefixes must start with '/'."
                    % cli_args.prefix)
            exit(1)
        site_prefix = cli_args.prefix
    elif cli_args.no_prefix:
        site_prefix = None

    broken_links, num_links_checked = check_directory(out_path, cli_args.offline, site_prefix)

    if not cli_args.no_final_retry and not cli_args.offline:
        final_retry_links(broken_links)
        #^ sleeps for FINAL_RETRY_DELAY and then retries remote links
        # Automatically removes from broken_links if they work now

    print("---------------------------------------")
    print("Link check report. %d links checked." % num_links_checked)

    if not cli_args.strict:
        unknown_broken_links = [ (page,link) for page,link in broken_links
                        if link not in config["known_broken_links"] ]

    if not broken_links:
        print("Success! No broken links found.")
    else:
        print("%d broken links found:"%(len(broken_links)))
        [print("File:",fname,"Link:",link) for fname,link in broken_links]

        if cli_args.strict or unknown_broken_links:
            exit(1)

        print("Success - all broken links are known problems.")


def dispatch_main():
    cli = DactylCLIParser(DactylCLIParser.UTIL_LINKS)
    global config
    config = DactylConfig(cli.cli_args)
    main(cli.cli_args)

if __name__ == "__main__":
    dispatch_main()



################################################################################
## Add version to markdown filter                                             ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Adds a message to the beginning of a file with a version number, based on  ##
## the URL of the remotely-fetched markdown.                                  ##
################################################################################
import re
import logging

def filter_markdown(md, currentpage={}, **kwargs):
    """Finds the version number and adds it to the start of the page."""
    version_regex = r"https://raw.githubusercontent.com/([A-Za-z0-9_.-]+)/([A-Za-z0-9_.-]+)/([A-Za-z0-9_-]+\.[A-Za-z0-9_.-]+)/.+\.md"

    try:
        version_match = re.match(version_regex, currentpage["md"])
    except (TypeError, KeyError):
        logging.warning("couldn't get MD path from page %s" % currentpage)
        return md

    try:
        github_owner = version_match.group(1)
        github_project = version_match.group(2)
        vnum = version_match.group(3)
        url = "https://github.com/%s/%s/releases/%s" % (github_owner, github_project, vnum)
        md = ("<p style='margin-top: 1em; font-style: italic'>Updated for <a href='%s' title='view on GitHub'>version %s</a></p>"%(url, vnum))+md
    except AttributeError:
        logging.warning("version regex didn't match: %s" % version_match)

    return md



################################################################################
## Remove doctoc filter                                                       ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Removes an automatically-generated "doctoc" table of contents, as          ##
## delineated by HTML comments, from the markdown source.                     ##
################################################################################


def filter_markdown(md, **kwargs):
    """Strip out doctoc Table of Contents for RippleAPI"""
    DOCTOC_START = "<!-- START doctoc generated TOC please keep comment here to allow auto update -->"
    DOCTOC_END = "<!-- END doctoc generated TOC please keep comment here to allow auto update -->"

    doctoc_start_i = md.find(DOCTOC_START)
    doctoc_end_i = md.find(DOCTOC_END)
    if doctoc_start_i != -1 and doctoc_end_i != -1:
        md = md[:doctoc_start_i]+md[doctoc_end_i+len(DOCTOC_END):]
    return md



################################################################################
## Buttonize links                                                            ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Looks for links ending in >, and adds a "button" class to those links so   ##
## they can be styled like buttons in the page.                               ##
################################################################################
import re

def filter_soup(soup, **kwargs):
    """make links ending in > render like buttons"""
    buttonlinks = soup.find_all("a", string=re.compile(">$"))
    for link in buttonlinks:
        link.string=link.string[:-1].strip()
        if "class" in link.attrs:
            link["class"].append("button")
        else:
            link["class"] = "button"



################################################################################
## Callouts filter                                                            ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Looks for sections starting **Note:** or **Caution:** and gives them CSS   ##
## classes like "callout note" so they can be styled accordingly.             ##
################################################################################
import re

CALLOUT_TYPES_FIELD = "callout_types"
CALLOUT_CLASS_FIELD = "callout_class"

DEFAULT_CALLOUT_TYPES = [
    "note",
    "warning",
    "tip",
    "caution"
]
DEFAULT_CALLOUT_CLASS = "dactyl-callout"

from bs4.element import Tag

def filter_soup(soup, currentpage={}, config={}, **kwargs):
    """
    Find patterns that look like callouts, for example **Note:**, and add
    callout classes to their parent elements (usually <p>)
    """
    # callout classes are defined by page>target>config>default
    callout_classes = currentpage.get(CALLOUT_TYPES_FIELD,
                        config.get(CALLOUT_TYPES_FIELD,
                        DEFAULT_CALLOUT_TYPES))
    callout_intro = re.compile(r"("+"|".join(callout_classes)+"):?$", re.I)
    callout_base_class = currentpage.get(CALLOUT_CLASS_FIELD,
                        config.get(CALLOUT_CLASS_FIELD,
                        DEFAULT_CALLOUT_CLASS))

    callouts = soup.find_all(name=["strong","em"], string=callout_intro)
    for c in callouts:
        if not c.previous_sibling: #This callout starts a block
            callout_type = c.string.replace(":","").lower()
            if callout_type in callout_classes:
                if (c.parent.parent.name == "blockquote" and Tag not in
                    [type(u) for u in c.parent.previous_siblings]):
                    # Special case for blockquotes, to allow multiline callouts.
                    # First element of BQ must start with a callout keyword
                    callout_el = c.parent.parent
                else:
                    callout_el = c.parent
                callout_el["class"] = [callout_base_class, callout_type]



################################################
## External Link Marker
## Author: Rome Reginelli
## Copyright: Ripple Labs, 2019
##
## Finds external links and changes them to:
## - open in a new tab (target="_blank")
## - end with an "external link" icon
##      (FontAwesome 4 required)
################################################

import re

def filter_soup(soup, **kwargs):
    """
    Adds an external link marker to external links
    and makes them open in new tabs.
    """

    extern_regex = re.compile(r"^https?://")

    links = soup.find_all("a", href=True)
    for link in links:
        if extern_regex.match(link["href"]):
            link["target"] = "_blank"
            ex_link_marker = soup.new_tag("i", attrs={
                    "class":"fa fa-external-link",
                    "aria-hidden": "true"})
            link.append(" ")
            link.append(ex_link_marker)
            oldclass = link.get('class',[])
            if type(oldclass) == str:
                oldclass = [oldclass]
            link['class'] = oldclass + ['external-link']



#!/usr/bin/env python3

################################################################################
# Dactyl - a tool for heroic epics of documentation
#
# Generates a website from Markdown and Jinja templates, with filtering
# along the way.
################################################################################

from dactyl.common import *

# Necessary to copy static files to the output dir
from distutils.dir_util import copy_tree, remove_tree
from shutil import copy as copy_file

# shallow copying of data types
from copy import copy

# Necessary for prince
import subprocess
from time import sleep

# Used to fetch markdown sources from the net
import requests
from urllib.parse import urlparse

# Various content and template processing stuff
import jinja2
from markdown import markdown
from bs4 import BeautifulSoup

from watchdog.observers import Observer

from dactyl.config import DactylConfig
from dactyl.cli import DactylCLIParser
from dactyl.target import DactylTarget
from dactyl.page import DactylPage
from dactyl.watch_handler import UpdaterHandler


class DactylBuilder:
    def __init__(self, target, config, mode="html", only_page=None):
        assert isinstance(target, DactylTarget)
        self.target = target
        self.config = config
        self.mode = mode
        # Generate a unique nonce per-run to be used for tempdir folder names
        self.nonce = str(time.time()).replace(".","")

        # Set a bunch of settings that can be overwritten after instantiating
        if mode == "html":
            self.copy_content_static = True
            self.copy_template_static = True
        elif mode == "md":
            self.copy_content_static = True
            self.copy_template_static = False
        else:
            self.copy_content_static = False
            self.copy_template_static = False

        self.out_path = self.config["out_path"]
        if mode == "pdf":
            self.staging_folder = self.temp_dir()
            self.pdf_filename = PDF_USE_DEFAULT

        self.leave_temp_files = False

        if mode == "es":
            self.es_upload = ES_USE_DEFAULT
        else:
            self.es_upload = NO_ES_UP

        if (self.config["template_allow_undefined"] == False and
                not self.config.bypass_errors):
            self.strict_undefined = True
        else:
            self.strict_undefined = False

        self.setup_html_env() # sets self.html_env

        self.default_pdf_template = self.get_template(self.config["default_pdf_template"])
        self.default_html_template = self.get_template(self.config["default_template"])
        self.default_es_template = self.get_es_template(self.config["default_es_template"])

        self.http_port = DEFAULT_SERVER_PORT


    def temp_dir(self):
        run_dir = os.path.join(self.config["temporary_files_path"],
                          "dactyl-"+self.nonce)
        if not os.path.isdir(run_dir):
            os.makedirs(run_dir)
        return run_dir

    def build_one(self, only_page):
        """
        Helper for building just a single page from the target.
        """
        self.build_all(only_page=only_page)

    @staticmethod
    def match_only_page(only_page, currentpage_data):
        """
        Returns a boolean indicating whether currentpage (object) matches the
        only_page parameter (string). Used for build_one()
        """
        if not only_page:
            return False
        if only_page[-5:] == ".html":
            if (currentpage_data["html"] == only_page or
                    os.path.basename(currentpage_data["html"]) == only_page):
                return True
        elif only_page[-3:] == ".md":
            if "md" in currentpage_data and (currentpage_data["md"] == only_page or
                    os.path.basename(currentpage_data["md"]) == only_page):
                return True
        return False

    def build_all(self, only_page=None):
        """
        Build and write all pages in the target, according to the set mode,
        and upload their entries to ElasticSearch if requested.
        """

        logger.info("loading pages in target...")
        pages = self.target.load_pages()
        logger.info("... done loading pages in target")

        # Set up context that gets passed to several build/render functions
        # as well as filters
        context = {
            "current_time": time.strftime(self.config["time_format"]), # Get time once only
            "config": self.config,
            "mode": self.mode,
            "target": self.target.data, # just data, for legacy compat
            "pages": [p.data for p in pages], # just data, for legacy compat
            "categories": self.target.categories(),
        }

        es_data = {}
        matched_only = False
        for page in pages:
            if only_page:
                if self.match_only_page(only_page, page.data):
                    matched_only = True
                else:
                    logger.debug("only_page mode: skipping page %s" % page)
                    continue

            logger.info("Building page: %s"%page)

            page_context = {"currentpage":page.data, **context}

            if self.mode == "es" or self.es_upload != NO_ES_UP:
                es_template = self.template_for_page(page, mode="es")
                es_json_s = page.es_json(es_template, page_context)
                es_page_id = self.target.name+"."+page.data["html"]
                es_data[es_page_id] = es_json_s

            if self.mode == "html" or self.mode == "pdf":
                use_template = self.template_for_page(page)
                logger.debug("use_template is: %s" % use_template)
                page_text = page.render(use_template, page_context)
            elif self.mode == "md":
                if "md" not in page.data and "__md_generator" not in page.data:
                    logger.info("... md mode: Skipping page (no md): %s" % page)
                    continue
                page_text = page.md_content(page_context)
            elif self.mode == "es":
                page_text = es_json_s
            else:
                exit("build_all() error: unknown mode: %s" % self.mode)

            if page_text:
                self.write_page(page_text, page.filepath(self.mode))
            else:
                logger.warning("not writing empty page '%s'"%page.data["name"])

        if only_page and not matched_only:
            exit("Didn't find requested 'only' page '%s'" % only_page)

        if self.es_upload != NO_ES_UP:
            self.upload_es(es_data)

        if self.mode == "pdf":
            self.assemble_pdf(only_page)


    def template_for_page(self, page, mode=None):
        """
        Return the preferred template for the given page and mode, based on
        inherited settings from the page/target and global default.
        """
        if mode is None:
            mode = self.mode

        fieldname, default_template = {
            "html": ("template", self.default_html_template),
            "pdf": ("pdf_template", self.default_pdf_template),
            "es": ("es_template", self.default_es_template),
            "md": (None, None) # Markdown doesn't use a template
        }[mode]

        if fieldname in page.data:
            return self.get_template(page.data[fieldname])
        else:
            return default_template

    def write_page(self, page_text, filepath):
        """
        Writes HTML/MD/ES JSON out to the filesystem.
        """
        if self.mode == "pdf":
            # only the final pdf goes to out_path
            base_folder = self.staging_folder
        else:
            base_folder = self.out_path

        # Join folders in case the filepath is not just a flat file
        out_folder = os.path.join(base_folder, os.path.dirname(filepath))
        if not os.path.isdir(out_folder):
            logger.info("creating output folder %s" % out_folder)
            os.makedirs(out_folder)
        fileout = os.path.join(base_folder, filepath)
        if os.path.isdir(fileout):
            logger.info("Writing index file for out-file '%s'" % fileout)
            fileout = fileout+"index.html"
        with open(fileout, "w", encoding="utf-8") as f:
            logger.info("writing to file: %s..." % fileout)
            f.write(page_text)

    def copy_static(self, template_static=None, content_static=None, out_path=None):
        """
        Copy static files to the output directory.
        """

        if template_static is None:
            template_static = self.copy_template_static
        if content_static is None:
            content_static = self.copy_content_static
        if out_path is None:
            out_path = self.config["out_path"]

        if template_static:
            template_static_src = self.config["template_static_path"]

            if os.path.isdir(template_static_src):
                template_static_dst = os.path.join(out_path,
                                           os.path.basename(template_static_src))
                copy_tree(template_static_src, template_static_dst)
            else:
                logger.warning(("Template static path '%s' doesn't exist; "+
                                "skipping.") % template_static_src)

        if content_static:
            if "content_static_path" in self.config:
                if type(self.config["content_static_path"]) == str:
                    content_static_srcs = [self.config["content_static_path"]]
                else:
                    content_static_srcs = self.config["content_static_path"]

                for content_static_src in content_static_srcs:
                    if os.path.isdir(content_static_src):
                        content_static_dst = os.path.join(out_path,
                                            os.path.basename(content_static_src))
                        copy_tree(content_static_src, content_static_dst)
                    elif os.path.isfile(content_static_src):
                        content_static_dst = os.path.join(out_path,
                                            os.path.dirname(content_static_src))
                        logger.debug("Copying single content_static_path file '%s'." %
                                content_static_src)
                        copy_file(content_static_src, content_static_dst)
                    else:
                        logger.warning("Content static path '%s' doesn't exist; skipping." %
                                        content_static_src)
            else:
                logger.debug("No content_static_path in conf; skipping copy")


    def setup_html_env(self):
        """
        Set up a Jinja env to load custom templates for HTML / HTML->PDF builds.
        """
        if self.strict_undefined:
            preferred_undefined = jinja2.StrictUndefined
        else:
            preferred_undefined = jinja2.ChainableUndefined

        loaderset = [jinja2.PackageLoader(__name__)]
        if "template_path" in self.config:
            loaderset.insert(0, jinja2.FileSystemLoader(self.config["template_path"]))
        env = jinja2.Environment(undefined=preferred_undefined,
                    extensions=['jinja2.ext.i18n'],
                    loader=jinja2.ChoiceLoader(loaderset))

        # Customize env: add custom tests, lstrip & trim blocks
        def defined_and_equalto(a,b):
            return env.tests["defined"](a) and env.tests["equalto"](a, b)
        env.tests["defined_and_equalto"] = defined_and_equalto
        def undefined_or_ne(a,b):
            return env.tests["undefined"](a) or env.tests["ne"](a, b)
        env.tests["undefined_or_ne"] = undefined_or_ne

        env.lstrip_blocks = True
        env.trim_blocks = True

        # Set up internationalization
        mo_file = self.target.data.get("locale_file", None)
        if mo_file:
            logger.debug("Loading strings from locale_file %s"%mo_file)
            try:
                with open(mo_file, "rb") as f:
                    tl = gettext.GNUTranslations(f)
                # TODO: add_fallback?? maybe a general config option there...
            except Exception as e:
                recoverable_error("Failed to load locale_file %s: %s" %
                                  (mo_file, e), self.config.bypass_errors,
                                  error=e)
                tl = gettext.NullTranslations()
        else:
            logger.debug("No locale_file setting found.")
            tl = gettext.NullTranslations()
        env.install_gettext_translations(tl, newstyle=True)

        self.html_env = env
        return env


    def get_template(self, template_name):
        """
        Gets the named Jinja template from the user-specified template folder,
        and falls back to the Dactyl built-in templates if it doesn't.
        """

        try:
            t = self.html_env.get_template(template_name)
        except jinja2.exceptions.TemplateError as e:
            recoverable_error("Error parsing template '%s': %s." %
                         (template_name, e), self.config.bypass_errors, error=e)
            t = jinja2.Template("Failed to load template '%s'."%template_name)
        return t

    def get_es_template(self, filename):
        """
        Loads an ElasticSearch template (as JSON)
        """
        try:
            template_path = os.path.join(self.config["template_path"], filename)
            with open(template_path, encoding="utf-8") as f:
                es_template = json.load(f)
        except (FileNotFoundError, json.decoder.JSONDecodeError, KeyError) as e:
            if type(e) == FileNotFoundError:
                logger.debug("Didn't find ES template (%s), falling back to default" %
                    template_path)
            elif type(e) == json.decoder.JSONDecodeError:
                recoverable_error(("Error JSON-decoding ES template (%s)" %
                    template_path), self.config.bypass_errors,
                    error=e)
            elif type(e) == KeyError:
                logger.debug("template_path isn't defined. No config file?")
            with resource_stream(__name__, BUILTIN_ES_TEMPLATE) as f:
                # es_template = json.load(f) # Doesn't work on Python 3.5
                es_template = json.loads(f.read().decode())
        return es_template

    def upload_es(self, data):
        """
        Upload to ElasticSearch. data is a dict mapping page IDs to JSON strings
        """
        es_base = self.cleanup_es_url()
        es_index = self.target.es_index_name()

        for id, json_s in data.items():
            doc_type="article"## TODO: other options?

            url = "{es_base}/{index}/{doc_type}/{id}".format(
                es_base=es_base,
                index=es_index.lower(), # ES index names must be lowercase
                doc_type=doc_type,
                id=id,
            )
            headers = {"Content-Type": "application/json"}
            logger.info("Uploading to ES: PUT %s" % url)
            r = requests.put(url, headers=headers, data=json_s)
            if r.status_code >= 400:
                recoverable_error("ES upload failed with error: '%s'" % r.text,
                        self.config.bypass_errors)

    def cleanup_es_url(self):
        """
        Do some post-processing on user-specified ES URLs to clean them up,
        or supply a default ES URL if one wasn't provided.
        """
        # Should not be called if not doing es upload
        assert self.es_upload != NO_ES_UP

        if self.es_upload == ES_USE_DEFAULT:
            es_base_url = self.config.get("elasticsearch", "http://localhost:9200")
        else:
            # URL was explicitly specified
            es_base_url = self.es_upload

        # Make sure it has an http:// or https:// prefix
        if es_base_url[:7] != "http://" and es_base_url[:8] != "https://":
            es_base_url = "https://"+es_base_url

        if es_base_url[-1] == "/": # Drop trailing slash
            es_base_url = es_base_url[:-1]

        logger.debug("ElasticSearch base URL is '%s'" % es_base_url)
        return es_base_url

    def assemble_pdf(self, only_page=None):
        """
        Use Prince to combine temporary HTML files into a PDF.
        Called at the end of build_all()
        """
        assert self.pdf_filename != NO_PDF

        if self.pdf_filename == PDF_USE_DEFAULT:
            # Not sure if this can change in watch mode
            pdf_filename = self.target.default_pdf_name()
        else:
            pdf_filename = self.pdf_filename

        # self.staging_folder should contain HTML files by now. Copy static
        # files there too, since Prince will need them
        self.copy_static(True, True, out_path=self.staging_folder)

        # Make sure the path we're going to write the PDF to exists
        if not os.path.isdir(self.out_path):
            logger.info("creating output folder %s" % self.out_path)
            os.makedirs(self.out_path)
        abs_pdf_path = os.path.abspath(os.path.join(self.out_path, pdf_filename))


        # Change dir to the tempfiles path. This helps Prince and also lets us
        # use os.path.isdir()
        old_cwd = os.getcwd()
        os.chdir(self.staging_folder)

        # Start preparing the prince command
        args = [self.config["prince_executable"], '--javascript', '-o', abs_pdf_path, '--no-warn-css']

        pages = self.target.pages
        if only_page:
            pages = [p for p in pages if self.match_only_page(only_page, p.data)][:1]
            if not len(pages):
                recoverable_error("Couldn't find 'only' page %s" % only_page,
                    self.config.bypass_errors)
                return

        # Each HTML output file in the target is another arg to prince
        if self.http_port:
            serverurl = "http://localhost:%d/" % self.http_port
        else:
            serverurl = "" # port 0: don't use the server
        for p in pages:
            phtml = p.data["html"]
            # if os.path.isdir(phtml):
            #     phtml = phtml+"index.html"
            args.append(serverurl+phtml)

        if self.http_port:
            # Start up an HTTP server for Prince. This helps it resolve absolute
            # URLs in hyperlinks correctly (which --fileroot doesn't quite do)
            server = subprocess.Popen(["python3", "-m", "http.server", str(self.http_port)])
            # Wait for server to start
            sleep(1)

        logger.info("generating PDF: running %s" % " ".join(args))
        prince_resp = subprocess.check_output(args, universal_newlines=True)
        print(prince_resp)

        if self.http_port:
            # Now shut down the server
            server.terminate()

        # Clean up the staging_folder now that we're done using it
        os.chdir(old_cwd)
        if not self.leave_temp_files:
            remove_tree(self.staging_folder)


    def watch(self):
        """
        Look for changed files and re-run the build whenever there's an update.
        Runs until interrupted.

        replacement for:
        watch(mode, target, cli_args.only, cli_args.pdf,
                es_upload=cli_args.es_upload,)
        """

        if self.mode == "html" and self.http_port:
            server = subprocess.Popen(["python3", "-m", "http.server",
                                      str(self.http_port), "-d", self.out_path])

        event_handler = UpdaterHandler(builder=self)
        observer = Observer()
        observer.schedule(event_handler, self.config["template_path"], recursive=True)
        observer.schedule(event_handler, self.config["content_path"], recursive=True)
        observer.start()
        # The above starts an observing thread,
        #   so the main thread can just wait
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

def list_targets(config):
    rows = []
    for t in config["targets"]:
        if "display_name" in t:
            display_name = t["display_name"]
        else:
            filename_field_vals = [t[f] for
                    f in config["pdf_filename_fields"]
                    if f in t.keys()]
            display_name = " ".join(filename_field_vals)
        rows.append((t["name"], display_name))

    col1_width = max([len(i) for i,j in rows]) + 3
    for row in rows:
        print("{row[0]:{width}} {row[1]}".format(row=row, width=col1_width))

def main(cli_args, config):
    if cli_args.list_targets_only:
        list_targets(config)
        exit(0)

    if cli_args.pages:
        config["content_path"] = "./" # page paths will be relative to CWD
        target = DactylTarget(config, inpages=cli_args.pages)
    elif cli_args.openapi:
        target = DactylTarget(config, spec_path=cli_args.openapi)
    else:
        target = DactylTarget(config, name=cli_args.target)

    if cli_args.vars:
        try:
            if cli_args.vars[-5:] in (".json",".yaml"):
                with open(cli_args.vars, "r", encoding="utf-8") as f:
                    custom_keys = yaml.load(f)
            else:
                custom_keys = yaml.load(cli_args.vars)
            target.gain_fields(custom_keys)
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            exit("FATAL: --vars value was improperly formatted: %s" % repr(e))

    if cli_args.title:
        target.gain_fields({"display_name": cli_args.title})

    if not cli_args.no_cover and not target.data.get("no_cover", False):
        target.add_cover()

    if cli_args.pdf != NO_PDF:
        mode = "pdf"
    elif cli_args.md:
        mode = "md"
    elif cli_args.es:
        mode = "es"
    else:
        mode = "html"

    builder = DactylBuilder(target, config, mode)

    if mode == "pdf":
        builder.pdf_filename = cli_args.pdf
    if mode == "es" or cli_args.es_upload:
        builder.es_upload = cli_args.es_upload

    # Override static files copy setting based on CLI flags, if any are used
    if cli_args.no_static:
        builder.copy_content_static = False
        builder.copy_template_static = False
    elif cli_args.template_static:
        builder.copy_content_static = False
        builder.copy_template_static = True
    elif cli_args.content_static:
        builder.copy_content_static = True
        builder.copy_template_static = False
    elif cli_args.copy_static:
        builder.copy_content_static = True
        builder.copy_template_static = True

    if cli_args.leave_temp_files:
        builder.leave_temp_files = True

    builder.http_port = cli_args.http_port

    if cli_args.only:
        logger.info("building page %s..."%cli_args.only)
        builder.build_one(cli_args.only)
    else:
        logger.info("building target %s..."%target.name)
        builder.build_all()
    logger.info("done building")

    builder.copy_static()

    if cli_args.watch:
        logger.info("watching for changes...")
        builder.watch()


def dispatch_main():
    cli = DactylCLIParser(DactylCLIParser.UTIL_BUILD)
    config = DactylConfig(cli.cli_args)
    config.load_build_options()
    main(cli.cli_args, config)


if __name__ == "__main__":
    dispatch_main()



################################################################################
## OpenAPI Specification Parsing
##
## Parses an OpenAPI spec in YAML or JSON and generates files from it
################################################################################

import jinja2
import json
import requests
import datetime
from urllib.parse import unquote as urldecode
from copy import deepcopy
from ruamel.yaml.comments import CommentedMap as YamlMap
from ruamel.yaml.comments import CommentedSeq as YamlSeq

from dactyl.common import *
from dactyl.http_constants import HTTP_METHODS, HTTP_STATUS_CODES, HTTP_METHODS_WITH_REQ_BODIES

DATA_TYPES_SUFFIX = "-data-types"
METHOD_TOC_SUFFIX = "-methods"

TOC_TEMPLATE = "template-openapi_endpoint_toc.md"
TAG_TOC_TEMPLATE = "template-openapi_endpoint_tag_toc.md"
ENDPOINT_TEMPLATE = "template-openapi_endpoint.md"
DATA_TYPES_TOC_TEMPLATE = "template-openapi_data_types_toc.md"
DATA_TYPE_TEMPLATE = "template-openapi_data_type.md"

class ApiDef:
    cached_specs = {}
    def __init__(self, spec_path, api_slug=None, extra_fields={},
                template_path=None):
        self.read_swag(spec_path)
        self.clean_up_swag()
        self.deref_swag()

        try:
            self.api_title = self.swag["info"]["title"]
        except IndexError:
            self.api_title = fname.replace(".yml","")+" API (working title)"

        if api_slug is None:
            self.api_slug = slugify(self.api_title)
        else:
            self.api_slug = api_slug

        self.extra_fields = extra_fields
        self.setup_jinja_env(template_path)

    @classmethod
    def from_path(cls, spec_path, api_slug=None, extra_fields={},
                template_path=None):
        """
        Instantiate an ApiDef instance only if we haven't done so already. This
        saves the trouble of fetching & parsing API specs more than once.
        """
        if spec_path in cls.cached_specs.keys():
            return cls.cached_specs[spec_path]

        apidef = cls(spec_path, api_slug=api_slug,
                extra_fields=extra_fields, template_path=template_path)
        cls.cached_specs[spec_path] = apidef
        return apidef

    def read_swag(self, spec_path):
        """Read the OpenAPI definition from either a local file or a URL, and
        store it at self.swag"""
        logger.debug("Reading OpenAPI definition from %s"%spec_path)

        if spec_path[:5] == "http:" or spec_path[:6] == "https:":
            response = requests.get(spec_path)
            if response.status_code == 200:
                self.swag = yaml.load(response.text)
            else:
                raise requests.RequestException("Status code for page was not 200")
        else:
            with open(spec_path, "r", encoding="utf-8") as f:
                self.swag = yaml.load(f)


    def setup_jinja_env(self, template_path=None):
        """Sets up the environment used to inject OpenAPI data into Markdown
        templates"""
        if template_path is None:
            loader = jinja2.PackageLoader(__name__)
        else:
            logger.debug("OpenAPI spec: preferring templates from %s"%template_path)
            loader = jinja2.ChoiceLoader([
                jinja2.FileSystemLoader(template_path),
                jinja2.PackageLoader(__name__)
            ])
        self.env = jinja2.Environment(loader=loader, extensions=['jinja2.ext.i18n'])
        self.env.lstrip_blocks = True
        self.env.rstrip_blocks = True

    @staticmethod
    def dig(parts, context):
        """
        Search a context object for something matching a $ref (recursive)
        """
        key = parts[0].replace("~1", "/").replace("~0", "~") # unescaped
        key = urldecode(key)
        try:
            key = int(key)
        except:
            pass
        if key not in context.keys():
            raise IndexError(key)

        if len(parts) == 1:
            return context[key]
        else:
            return ApiDef.dig(parts[1:], context[key])

    def deref(self, ref):
        """Look through the YAML for a specific reference key, and return
        the value that key represents.
        - Raises IndexError if the key isn't found
            in the YAML.
        - add_title: If true, provide a "title" field when the reference
            resolves to an object that doesn't have a "title". The provided
            "title" value is based on the key that contained the reference
        """
        assert len(ref) > 1 and ref[0] == "#" and ref[1] == "/"
        parts = ref[2:].split("/")
        assert len(parts) > 0

        return self.dig(parts, self.swag)

    def deref_swag(self):
        """
        Walk the OpenAPI specification for $ref objects and resolve them to
        the values they reference. Assumes the entire spec is contained in a
        single file.
        """

        def deref_yaml(yaml_value):
            if "keys" in dir(yaml_value): # Dictionary-like type
                if "$ref" in yaml_value.keys():
                    # It's a reference; deref it
                    reffed_value = self.deref(yaml_value["$ref"])
                    # The referenced object may contain more references, so
                    # resolve those before returning
                    return deref_yaml(reffed_value)
                else:
                    # recurse through each key/value pair looking for refs
                    the_copy = YamlMap()
                    for k,v in yaml_value.items():
                        the_copy[k] = deref_yaml(v)
                    return the_copy
            elif "append" in dir(yaml_value): # List-like type
                # recurse through each item looking for refs
                the_copy = YamlSeq()
                for item in yaml_value:
                    the_copy.append(deref_yaml(item))
                return the_copy
            else: # Probably a basic type
                # base case: return the value
                return yaml_value

        self.swag = deref_yaml(self.swag)

    def clean_up_swag(self):
        # Give each schema in the "components" a title if it's missing one
        schemas = self.swag.get("components", {}).get("schemas", {})
        for key,schema in schemas.items():
            title = schema.get("title", key)
            schema["title"] = title
            if "example" in schema:
                try:
                    j = json.dumps(schema["example"], indent=4, default=self.json_default)
                    schema["example"] = j
                except Exception as e:
                    logger.debug("%s example isn't json: %s"%(title,j))
        self.deref_swag()

        # Find all tags used in endpoints and add any undefined ones to the
        # top level OpenAPI Object
        taglist = self.swag.get("tags", [])
        for path, method, endpoint in self.endpoint_iter():
            etags = endpoint.get("tags", [])
            for tag in etags:
                if tag not in [t.get("name") for t in taglist]:
                    taglist.append({
                        "name": tag,
                    })
        self.swag["tags"] = taglist


    def render_method_toc(self):
        t = self.env.get_template(TOC_TEMPLATE)
        context = self.new_context()
        context["endpoints"] = self.endpoint_iter()
        context["endpoints_by_tag"] = self.endpoint_iter
        return t.render(self.swag, **context)

    def render_tag_toc(self, tag):
        t = self.env.get_template(TAG_TOC_TEMPLATE)
        context = self.new_context()
        context["tag"] = tag
        context["endpoints"] = self.endpoint_iter()
        context["endpoints_by_tag"] = self.endpoint_iter
        return t.render(self.swag, **context)

    def render_data_types_toc(self):
        t = self.env.get_template(DATA_TYPES_TOC_TEMPLATE)
        context = self.new_context()
        context["schemas"] = self.data_type_iter()
        return t.render(self.swag, **context)

    def render_data_type(self, key, schema):
        t = self.env.get_template(DATA_TYPE_TEMPLATE)
        context = self.new_context()
        if "title" not in schema.keys():
            schema["title"] = key
        return t.render(schema, **context)

    def render_endpoint(self, path, method, endpoint):
        t = self.env.get_template(ENDPOINT_TEMPLATE)
        context = self.new_context()
        context["method"] = method
        context["path"] = path
        context["path_params"] = [p for p in endpoint.get("parameters",[]) if p["in"]=="path"]
        context["query_params"] = [p for p in endpoint.get("parameters",[]) if p["in"]=="query"]
        context["x_example_request_body"] = self.get_x_example_request_body(path,method,endpoint)
        #TODO: header & cookie params?? example response body?
        return t.render(endpoint, **context)

    def get_x_example_request_body(self, path, method, endpoint):
        if method not in HTTP_METHODS_WITH_REQ_BODIES:
            return ""

        content = endpoint.get("requestBody",{}).get("content",{})
        if not content:
            return ""
        for mediatype,content_inner in content.items():
            if "example" in content_inner:
                # single example
                ex = content_inner["example"]
            else:
                try:
                    # multiple examples? use the first one
                    ex = list(content_inner["examples"].values())[0]["value"]
                except (IndexError, KeyError, AttributeError) as e:
                    logger.debug("Media type %s didn't have an example value"%mediatype)
                    return ""

            try:
                ex_pp = json.dumps(ex, indent=4, separators=(',', ': '), default=self.json_default)
            except TypeError as e:
                traceback.print_tb(e.__traceback__)
                logger.debug("json dumps failed on example '%s'"%ex)
                ex_pp = ex
            return ex_pp

        logger.debug("couldn't find an example value for %s %s"%(method,path))
        return ""

    def get_endpoint_renderer(self, path, method, endpoint):
        return lambda: self.render_endpoint(path, method, endpoint)

    def get_data_type_renderer(self, key, schema):
        return lambda: self.render_data_type(key, schema)

    def get_tag_toc_renderer(self, tag):
        return lambda: self.render_tag_toc(tag)

    def endpoint_iter(self, tag=None):
        paths = self.swag.get("paths", {})
        for path, path_def in paths.items():
            for method in HTTP_METHODS:
                if method in path_def.keys():
                    endpoint = path_def[method]
                    if tag==None or tag in endpoint.get("tags", []) or \
                            (tag=="Uncategorized" and endpoint.get("tags", []) == []):
                        # TODO: Inherit parameters from the path definition itself
                        # Fill in some "sensible defaults" for fields we really want
                        operationId = endpoint.get("operationId", slugify(method+path))
                        endpoint["operationId"] = operationId
                        summary = endpoint.get("summary", operationId)
                        endpoint["summary"] = summary
                        yield (path, method, endpoint)

    def data_type_iter(self):
        schemas = self.swag.get("components", {}).get("schemas", {})
        for key,schema in schemas.items():
            title = schema.get("title", key)
            schema["title"] = title
            yield (title, schema)

    def create_pagelist(self):
        """
        Return an array of pages representing this API, which Dactyl can use
        as it would use a normal list of pages in the config
        """
        pages = []

        # TODO: make all the blurb/category strings template strings that can
        #       be translated and configured

        # add methods table of contents
        toc_page = deepcopy(self.extra_fields)
        toc_page.update({
            "name": "All "+self.api_title+" Methods",
            "__md_generator": self.render_method_toc,
            "html": self.api_slug+METHOD_TOC_SUFFIX+".html",
            "blurb": "List of methods/endpoints available in "+self.api_title,
            "category": "All "+self.api_title+" Methods",
        })
        if "parent" not in toc_page:
            toc_page["parent"] = "index.html"
        pages.append(toc_page)

        # add a table of contents per tag
        for tag in self.swag.get("tags",[{"name": "Uncategorized","description":""}]):
            tag_toc_page = deepcopy(self.extra_fields)
            tag_toc_page.update({
                "name": tag["name"].title()+" Methods",
                "__md_generator": self.get_tag_toc_renderer(tag),
                "html": self.api_slug+"-"+tag["name"]+METHOD_TOC_SUFFIX+".html",
                "blurb": tag.get("description",""),
                "category": tag["name"].title()+" Methods",
                "parent": toc_page["html"],
            })
            pages.append(tag_toc_page)

            # Add endpoints for this tag, except duplicates
            for path, method, endpoint in self.endpoint_iter(tag["name"]):
                tag0 = endpoint.get("tags",["Uncategorized"])[0]
                if tag0 != tag["name"]:
                    continue # Skip method whose primary tag is not this one
                method_page = deepcopy(self.extra_fields)
                method_page.update({
                    "name": endpoint["summary"],
                    "__md_generator": self.get_endpoint_renderer(path, method, endpoint),
                    "html": self.method_link(path, method, endpoint),
                    "blurb": endpoint.get("description", endpoint["operationId"]+" method"),
                    "category": tag0+" Methods",
                    "parent": tag_toc_page["html"],
                })
                pages.append(method_page)

        # add data types table of contents
        data_types_page = deepcopy(self.extra_fields)
        data_types_page.update({
            "name": self.api_title+" Data Types",
            "__md_generator": self.render_data_types_toc,
            "html": self.api_slug+DATA_TYPES_SUFFIX+".html",
            "blurb": "List of all data types defined for "+self.api_title,
            "category": self.api_title+" Data Types",
            "parent": toc_page["html"],
        })
        pages.append(data_types_page)

        # add each data type from the components.schemas list
        schemas = self.swag.get("components", {}).get("schemas", {})
        for title, schema in self.data_type_iter():
            data_type_page = deepcopy(self.extra_fields)
            data_type_page.update({
                "name": title,
                "__md_generator": self.get_data_type_renderer(title, schema),
                "html": self.type_link(title),
                "blurb": "Definition of "+title+" data type",
                "category": self.api_title+" Data Types",
                "parent": data_types_page["html"],
            })
            pages.append(data_type_page)

        return pages

    def add_metadata(self, target_data):
        """
        Extend the provided target_data dictionary with metadata pulled from
        the spec.
        """
        info = self.swag.get("info", {"title":self.api_title, "version":"0.0.0"})
        target_data["info"] = info


    def new_context(self):
        return {
            "api_title": self.api_title,
            "type_link": self.type_link,
            "method_link": self.method_link,
            "HTTP_METHODS": HTTP_METHODS,
            "HTTP_STATUS_CODES": HTTP_STATUS_CODES,
            "spec": self.swag,
            "debug": logger.debug,
            "slugify": slugify,
            "md_escape": self.md_escape,
        }

    @staticmethod
    def md_escape(text):
        """
        Escape potential Markdown syntax in a string.
        This is meant to be passed to the templates.
        """
        specialchars = "\\`*_{}[]()#+-.!"
        s = ""
        for c in text:
            if c in specialchars:
                s += "\\"
            s += c
        return s

    @staticmethod
    def json_default(o):
        """
        Serializer function for JSON (from YAML)
        """
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()
        else:
            return str(o)

    def type_link(self, title):
        # TODO: in "md" mode, use ".md" suffix
        return self.api_slug+DATA_TYPES_SUFFIX+"-"+slugify(title.lower())+".html"

    def method_link(self, path, method, endpoint):
        return self.api_slug+"-"+slugify(endpoint["operationId"]+".html")



################################################################################
## Standardize header IDs                                                     ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Replaces underscores with dashes in h1,h2,... element IDs. This provides   ##
## compatibility with some other flavors of markdown that generate HTML IDs   ##
## differently.                                                               ##
################################################################################
import re

def filter_soup(soup, **kwargs):
    """replace underscores with dashes in h1,h2,etc. for backwards compatibility"""
    headers = soup.find_all(name=re.compile("h[0-9]"), id=True)
    for h in headers:
        if "_" in h["id"]:
            h["id"] = h["id"].replace("_", "-")



################################################################################
## Dactyl Page Class
##
## Handles the loading, default values, preprocessing, and content parsing
## for a single HTML page of output.
################################################################################

import jinja2
import requests

from markdown import markdown
from bs4 import BeautifulSoup

from dactyl.common import *

from dactyl.jinja_loaders import FrontMatterRemoteLoader, FrontMatterFSLoader

class DactylPage:
    def __init__(self, config, data, skip_pp=False, more_filters=[]):
        self.config = config
        self.data = data
        self.rawtext = None
        self.pp_template = None
        self.twolines = None
        self.toc = []
        self.ffp = None
        self.skip_pp = skip_pp
        self.md = None
        self.html = None
        self.soup = None

        self.gain_filters(more_filters)
        self.load_content()
        self.provide_default_filename()
        self.provide_name()
        # self.html_content({"currentpage":data, **context})

    def get_pp_env(self, loader):
        if (self.config["preprocessor_allow_undefined"] or
                self.config.bypass_errors):
            preferred_undefined = jinja2.Undefined
        else:
            preferred_undefined = jinja2.StrictUndefined

        pp_env = jinja2.Environment(undefined=preferred_undefined,
                extensions=['jinja2.ext.i18n'], loader=loader)

        # Add custom "defined_and_" tests
        def defined_and_equalto(a,b):
            return pp_env.tests["defined"](a) and pp_env.tests["equalto"](a, b)
        pp_env.tests["defined_and_equalto"] = defined_and_equalto
        def undefined_or_ne(a,b):
            return pp_env.tests["undefined"](a) or pp_env.tests["ne"](a, b)
        pp_env.tests["undefined_or_ne"] = undefined_or_ne

        # Pull exported values (& functions) from page filters into the pp_env
        for filter_name in self.filters(save=False):
            if filter_name not in self.config.filters.keys():
                logger.debug("Skipping unloaded filter '%s'" % filter_name)
                continue
            if "export" in dir(self.config.filters[filter_name]):
                for key,val in self.config.filters[filter_name].export.items():
                    logger.debug("... pulling in filter_%s's exported key '%s'"
                            % (filter_name, key))
                    pp_env.globals[key] = val

        return pp_env

    def load_content(self):
        """
        Dispatcher for loading this page's content based on the "md" field.
        """
        logger.debug("Loading page %s" % self)
        if "md" in self.data:
            if (self.data["md"][:5] == "http:" or
                    self.data["md"][:6] == "https:"):
                self.load_from_url()
            else:
                self.load_from_disk()
        elif "__md_generator" in self.data:
            self.load_from_generator()

    def load_from_url(self):
        """
        Read file over HTTP(S),
        as either raw text or as a Jinja template,
        and load frontmatter, if any, either way.
        """
        url = self.data["md"]
        logger.info("Loading page from URL: %s"%url)
        assert (url[:5] == "http:" or url[:6] == "https:")
        if not self.skip_pp:
            pp_env = self.get_pp_env(loader=FrontMatterRemoteLoader())
            self.pp_template = pp_env.get_template(self.data["md"])
            frontmatter = pp_env.loader.fm_map[self.data["md"]]
            merge_dicts(frontmatter, self.data)
            # special case: let frontmatter overwrite default "html" vals
            if PROVIDED_FILENAME_KEY in self.data and "html" in frontmatter:
                self.data["html"] = frontmatter["html"]
            self.twolines = pp_env.loader.twolines[self.data["md"]]
        else:
            response = requests.get(url)
            if response.status_code == 200:
                self.rawtext, frontmatter = parse_frontmatter(response.text)
                merge_dicts(frontmatter, self.data)
                # special case: let frontmatter overwrite default "html" vals
                if PROVIDED_FILENAME_KEY in self.data and "html" in frontmatter:
                    self.data["html"] = frontmatter["html"]
                self.twolines = self.rawtext.split("\n", 2)[:2]
            else:
                raise requests.RequestException("Status code for page was not 200")

    def load_from_disk(self):
        """
        Read the file from the filesystem,
        as either raw text or as a Jinja template,
        and load frontmatter, if any, either way.
        """
        assert "md" in self.data
        if not self.skip_pp:
            logger.debug("... loading markdown from filesystem")
            path = self.config["content_path"]
            pp_env = self.get_pp_env(loader=FrontMatterFSLoader(path))
            self.pp_template = pp_env.get_template(self.data["md"])
            frontmatter = pp_env.loader.fm_map[self.data["md"]]
            merge_dicts(frontmatter, self.data)
            # special case: let frontmatter overwrite default "html" vals
            if PROVIDED_FILENAME_KEY in self.data and "html" in frontmatter:
                self.data["html"] = frontmatter["html"]
            self.twolines = pp_env.loader.twolines[self.data["md"]]
        else:
            logger.info("... reading markdown from file")
            fullpath = os.path.join(self.config["content_path"], self.data["md"])
            with open(fullpath, "r", encoding="utf-8") as f:
                ftext = f.read()
            self.rawtext, frontmatter = parse_frontmatter(ftext)
            merge_dicts(frontmatter, self.data)
            # special case: let frontmatter overwrite default "html" vals
            if PROVIDED_FILENAME_KEY in self.data and "html" in frontmatter:
                self.data["html"] = frontmatter["html"]
            self.twolines = self.rawtext.split("\n", 2)[:2]


    def load_from_generator(self):
        """
        Load the text from a generator function,
        as either raw text or a jinja template.
        Assume no frontmatter in this case.
        """
        if not self.skip_pp:
            pp_env = self.get_pp_env(
                loader=jinja2.DictLoader({"_": self.data["__md_generator"]()}) )
            self.pp_template = pp_env.get_template("_")
        else:
            self.rawtext = self.data["__md_generator"]()

    def provide_default_filename(self):
        """
        Provide a default "html" filename to a page dictionary if one wasn't
        explicitly specified. Frontmatter can overwrite this value, but code
        won't see the frontmatter-provided "html" value when not building the
        target containing that page.
        """
        if "html" in self.data:
            return

        logger.debug("Need to generate html filename for page %s" % self)
        if "md" in self.data:
            # TODO: support other formulas including "tail" or "pretty"
            new_filename = re.sub(r"[.]md$", ".html", self.data["md"])
            if self.config.get("flatten_default_html_paths", True):
                new_filename = new_filename.replace(os.sep, "-")
            self.data["html"] = new_filename
        elif "name" in self.data:
            new_filename = slugify(self.data["name"]).lower()+".html"
            self.data["html"] = new_filename
        else:
            new_filename = str(time.time()).replace(".", "-")+".html"
            self.data["html"] = new_filename
        self.data[PROVIDED_FILENAME_KEY] = True

        logger.debug("Generated html filename '%s' for page: %s" %
                    (new_filename, self))

    def provide_name(self):
        """
        Add the "name" field, if not defined.
        """
        if "name" in self.data:
            return

        if "title" in self.data: # Port over the "title" attribute instead
            self.data["name"] = self.data["title"]
            logger.debug("Guessed page name from title for page %s" % self)
            return
        elif self.twolines:
            logger.debug("Guessing page name from first two lines...")
            try:
                soup = BeautifulSoup(markdown("\n".join(self.twolines)), "html.parser")
                first_h = soup.find(name=re.compile("h[1-6]"))
                self.data["name"] = first_h.get_text()
                logger.debug("... guessed title: '%s'"%self.data["name"])
                return
            except Exception as e:
                traceback.print_tb(e.__traceback__)
                logger.warning(("Couldn't guess title of page '%s' from its "+
                                "first two lines:\n%s") % (self,
                                    "\n".join(self.twolines)))

        if "md" in self.data:
            self.data["name"] = self.data["md"]
            logger.debug("Using placeholder name for page based on md path: '%s'"%self.data["name"])
        else:
            logger.warning("Using a placeholder name for page %s" %
                    str(self.data))
            self.data["name"] = str(time.time()).replace(".", "-")



    def preprocess(self, context):
        try:
            md = self.pp_template.render(**context)
        except Exception as e:
            recoverable_error("Preprocessor error in page %s: %s."%(self, e),
                              self.config.bypass_errors, error=e)
            # Preprocessing failed. Skip this step and reload.
            self.skip_pp = True
            self.load_content()
            return self.md_content(context)
        # Apply markdown-based filters here
        for filter_name in self.filters():
            if "filter_markdown" in dir(self.config.filters[filter_name]):
                logger.debug("... applying markdown filter %s" % filter_name)
                try:
                    md = self.config.filters[filter_name].filter_markdown(
                        md,
                        logger=logger,
                        **context,
                    )
                except Exception as e:
                    recoverable_error("Markdown filter '%s' failed on page %s: %s" %
                                  (filter_name, self, e),
                                  self.config.bypass_errors, error=e)

        logger.info("... markdown is ready")
        self.md = md
        return md

    def md_content(self, context):
        if self.md is not None:
            # return already-preprocessed md
            return self.md
        elif self.rawtext is not None:
            return self.rawtext
        elif not self.skip_pp and self.pp_template is not None:
            return self.preprocess(context)
        else:
            logger.debug("page %s has no rawtext or pp_template"%self.data)
            return ""

    def html_content(self, context, regen=False, save=True):
        """
        Returns the page's contents as HTML. Parses Markdown & runs filters
        if any.
        """
        if self.html is not None and not regen:
            # Reuse saved results if we have them
            return self.html

        md = self.md_content(context)

        if md:
            logger.info("... parsing markdown...")
            html = markdown(md, extensions=["markdown.extensions.extra",
                                            "markdown.extensions.sane_lists"])
        else:
            html = ""

        # Apply raw-HTML-string-based filters here
        for filter_name in self.filters():
            if "filter_html" in dir(self.config.filters[filter_name]):
                logger.debug("... applying HTML filter %s" % filter_name)
                try:
                    html = self.config.filters[filter_name].filter_html(
                            html,
                            logger=logger,
                            **context,
                    )
                except Exception as e:
                    recoverable_error("HTML filter '%s' failed on page %s: %s" %
                            (filter_name, self, e), self.config.bypass_errors,
                            error=e)

        # Some filters would rather operate on a soup than a string.
        # May as well parse once and re-serialize once.
        soup = BeautifulSoup(html, "html.parser")
        self.soup = soup

        # Give each header a unique ID and fill out the Table of Contents
        self.update_toc()

        # Add a "blurb" attribute to the page
        self.provide_blurb()

        # Add this page's plaintext field. ElasticSearch upload uses this.
        self.data["plaintext"] = soup.get_text()

        # Apply soup-based filters here
        for filter_name in self.filters():
            if "filter_soup" in dir(self.config.filters[filter_name]):
                logger.debug("... applying soup filter %s" % filter_name)
                try:
                    self.config.filters[filter_name].filter_soup(
                            soup,
                            logger=logger,
                            **context,
                    )
                    # ^ the soup filters apply to the same object, passed by reference
                except Exception as e:
                    recoverable_error("Soup filter '%s' failed on page %s: %s" %
                                  (filter_name, self, e),
                                  self.config.bypass_errors, error=e)

        logger.info("... re-rendering HTML from soup...")
        html2 = str(soup)
        if save:
            self.html = html2
        return html2

    @staticmethod
    def idify(utext):
        """Make a string ID-friendly (but more unicode-friendly)"""
        utext = re.sub(r'[^\w\s-]', '', utext).strip().lower()
        utext = re.sub(r'[\s-]+', '-', utext)
        if not len(utext):
            # Headers must be non-empty
            return '_'
        return utext

    def update_toc(self):
        """
        Assign unique IDs to header elements in a BeautifulSoup object, and
        update internal table of contents accordingly.
        The resulting ToC is a list of objects, each in the form:
        {
            "text": "Header Content as Text",
            "id": "header-content-as-text", # doesn't have the # prefix
            "level": 1, #1-6, based on h1, h2, etc.
        }
        Also add the "headermap" field used for ES JSON, which is in the form
        {
            "Header Content as Text": "#header-content-as-text"
        }
        """

        self.toc = []
        uniqIDs = {}
        headermap = {}
        headers = self.soup.find_all(name=re.compile("h[1-6]"))
        for h in headers:
            h_id = self.idify(h.get_text())
            if h_id not in uniqIDs.keys():
                uniqIDs[h_id] = 0
            else:
                # not unique, append -1, -2, etc. to this instance
                uniqIDs[h_id] += 1
                h_id = "{id}-{n}".format(id=h_id, n=uniqIDs[h_id])

            h["id"] = h_id
            self.toc.append({
                "text": h.get_text(),
                "id": h_id,
                "level": int(h.name[1])
            })
            # ElasticSearch doesn't like dots in keys, so escape those
            escaped_name = h.get_text().replace(".","-")
            headermap[escaped_name] = "#"+h_id
        self.data["headermap"] = headermap

    def legacy_toc(self):
        """
        Return an HTML table of contents in the legacy format from the internal
        table of contents list.
        """
        soup = BeautifulSoup("", "html.parser")
        for h in self.toc:
            if h["level"] > 3:
                # legacy toc only goes down to h3
                continue
            a = soup.new_tag("a", href="#"+h["id"])
            a.string = h["text"]
            li = soup.new_tag("li")
            li["class"] = "level-{n}".format(n=h["level"])
            li.append(a)
            soup.append(li)
            soup.append("\n")
        return str(soup)

    def provide_blurb(self):
        """
        Add a "blurb" field, based on the first paragraph in the page, if one
        is not already provided.
        """
        if "blurb" in self.data:
            return
        p = self.soup.find("p")
        while p:
            if p.get_text().strip():
                self.data["blurb"] = p.get_text()
                break
            else:
                p = p.find_next_sibling("p")
        else:
            logger.debug("Couldn't find a paragraph with text in page %s." % self.data["name"])
            # No text? No blurb.
            self.data["blurb"] = ""

    def render(self, use_template, context):
        """
        Render the entire page using the given template & context.
        """
        # TODO: try block around html_content()?
        html_content = self.html_content(context)

        out_html = use_template.render(
            content=html_content,
            sidebar_content=self.legacy_toc(),
            page_toc=self.legacy_toc(),
            headers=self.toc,
            **context,
        )
        return out_html

    def es_json(self, use_template, context):
        """
        Return JSON for uploading to ElasticSearch
        """
        # Parse HTML to get the required "bonus" fields like blurb and toc, but
        # don't save the "es"-rendered HTML as the final output, in case this
        # is a build-HTML-and-upload-ES run.
        es_context = {**context}
        es_context["mode"] = "es"
        self.html_content(es_context, save=False)

        # Set up Jinja env for rendering strings
        es_env = self.get_pp_env(loader=None)#TODO: does this even work?

        def eval_es_string(expr):
            try:
                result = eval(expr, {}, context)
            except Exception as e:
                recoverable_error(
                    "__dactyl_eval__ failed on expression '{expr}': {e}".format(
                        expr=expr, e=e),
                    self.config.bypass_errors
                )
                result = expr
            return result

        def render_es_field(value):
            if type(value) == str: # jinja-render strings
                field_templ = es_env.from_string(value)
                try:
                    parsed_field = field_templ.render(context)
                except jinja2.exceptions.TemplateSyntaxError as e:
                    recoverable_error(
                        "Couldn't parse value '{value}' in ES template: {e}".format(
                            value, e),
                        self.config.bypass_errors
                    )
                return parsed_field
            elif type(value) in (type(None), int, float, bool): # preserve literals
                return value
            elif type(value) == dict and ES_EVAL_KEY in value.keys():
                return eval_es_string(value[ES_EVAL_KEY])
            elif type(value) == dict: # recurse!
                return {rkey: render_es_field(rval) for rkey,rval in value.items()}
            elif type(value) == list: # recurse!
                return [render_es_field(rval) for rval in value]
            else:
                recoverable_error("Unknown type '{t}' in ES template" % type(value))
                return ""

        wout = {key: render_es_field(val) for key,val in use_template.items()}
        return json.dumps(wout, indent=4, separators=(',', ': '))

    def filepath(self, mode):
        """
        Returns the preferred filename to write output to based on the provided
        mode.
        """

        if mode == "es":
            # use .json as file extension instead
            fp = re.sub(r'(.+)\.html?$', r'\1.json', self.data["html"], flags=re.I)
            if fp[-5:] != ".json": # substitution didn't work
                fp = fp+".json"
            return fp
        elif mode == "md":
            if "md" in self.data:
                # reuse the input .md filename as output
                fp = self.data["md"]
                if ":" in fp: # http: or https: probably
                    fp = slugify(f)
            else:
                # use the html field, but change the file extension to .md
                fp = re.sub(r'(.+)\.html?$', r'\1.md', self.data["html"], flags=re.I)
                if fp[-3:] != ".md": # substitution didn't work
                    fp = fp+".md"
            return fp
        else:
            # for pdf or html just use the html field as-is
            return self.data["html"]

    def gain_filters(self, filterlist):
        """
        Called by target to add its filters to this page's.
        """
        if "filters" in self.data:
            self.data["filters"] = filterlist + self.data["filters"]
        else:
            self.data["filters"] = filterlist

    def filters(self, save=True):
        """
        Returns the names of filters to use when processing this page.
        """
        if self.ffp:
            # Return saved results
            return self.ffp

        # Create de-duplicated list of loaded filters only
        ffp = []
        for f in self.config["default_filters"]:
            if f not in ffp and f in self.config.filters:
                ffp.append(f)
        for f in self.data.get("filters", [])    :
            if f not in ffp and f in self.config.filters:
                ffp.append(f)

        logger.debug("...filters for page: %s"%ffp)
        if save:
            self.ffp = ffp
        return ffp

    def __repr__(self):
        if "name" in self.data and "html" in self.data:
            return "<Page '{name}' ({html})>".format(
                name=self.data["name"],
                html=self.data["html"],
            )
        if "name" in self.data:
            return "<Page '{name}'>".format(name=self.data["name"])
        if "html" in self.data:
            return "<Page ({html})>".format(html=self.data["html"])
        strdata = str(self.data)
        if len(strdata) > 30:
            strdata = strdata[:27]+"..."
        return "<Page {{{strdata}}}>".format(strdata=strdata)



#!/usr/bin/env python3

###############################################################################
## Dactyl Style Police                                                       ##
## Author: Rome Reginelli                                                    ##
## Copyright: Ripple Labs, Inc. 2016                                         ##
##                                                                           ##
## Reads the markdown files to try and enforce elements of good style.       ##
###############################################################################

from dactyl.common import *

#import nltk
import collections

from bs4 import BeautifulSoup
from bs4 import Comment
from bs4 import NavigableString

from dactyl.dactyl_build import DactylBuilder
from dactyl.config import DactylConfig
from dactyl.cli import DactylCLIParser
from dactyl.target import DactylTarget

OVERRIDE_COMMENT_REGEX = r" *STYLE_OVERRIDE: *([\w, -]+)"

class DactylStyleChecker:
    def __init__(self, target, config):
        """
        TODO
        """
        self.target = target
        self.config = config
        self.load_style_rules()

    @staticmethod
    def tokenize(passage):
        words = re.split(r"[\s,.;()!'\"]+", passage)
        return [w for w in words if w]

    @staticmethod
    def depunctuate(passage):
        punctuation = re.compile(r"[,.;()!'\"]")
        return re.sub(punctuation, "", passage)

    def load_style_rules(self):
        """Reads word and phrase substitution files into the config"""
        if "word_substitutions_file" in self.config:
            with open(self.config["word_substitutions_file"], "r", encoding="utf-8") as f:
                self.disallowed_words = yaml.load(f)
        else:
            logger.warning("No 'word_substitutions_file' found in config.")
            self.disallowed_words = {}

        if "phrase_substitutions_file" in self.config:
            with open(self.config["phrase_substitutions_file"], "r", encoding="utf-8") as f:
                self.disallowed_phrases = yaml.load(f)
        else:
            logger.warning("No 'phrase_substitutions_file' found in config.")
            self.disallowed_phrases = {}

    def check_all(self):
        """
        Reads all pages for this checker's target and checks them for style.

        Sets self.issues to an array of tuples with any recommended changes.
        """

        logger.info("Loading pages in target %s"%self.target.name)
        pages = self.target.load_pages()
        logger.info("Done loading pages.")

        context = {
            "current_time": time.strftime(self.config["time_format"]), # Get time once only
            "config": self.config,
            "mode": "html",
            "target": self.target.data,
            "pages": [p.data for p in pages], # just data, for legacy compat
            "categories": self.target.categories(),
        }

        logger.info("Style Checker - checking all pages in target %s"%self.target.name)

        style_issues = []
        for page in pages:
            logger.info("Checking page %s..." % page)
            page_context = {"currentpage":page.data, **context}
            page.html_content(page_context) # This defines page.soup
            logger.debug("page.soup is... %s"%page.soup)
            # If the page has no content, soup can be empty and that's OK.
            if not page.soup.find_all():
                logger.info("...page %s has no content."%page)
                continue

            page_issues = []
            # All text in the parsed Markdown content should be contained in
            # one of these top-level elements, except we ignore code samples.
            block_elements = ["p","ul","table","h1","h2","h3","h4","h5","h6"]
            blocks = [el for el in page.soup.find_all(
                        name=block_elements, recursive=False)]

            for block in blocks:
                overrides = self.get_overrides(block)
                # "Wipe" inlined <code> elements so we don't style-check
                # code samples.
                [code.clear() for code in block.find_all("code")]
                passage_issues = self.check_passage(block.get_text(), overrides)
                if passage_issues:
                    page_issues += passage_issues

            if page_issues:
                style_issues.append( (page.data["name"], page_issues) )

        self.issues = style_issues
        return style_issues


    def report(self):
        """
        Print a report of discovered style issues to stdout.
        self.check_all() must be called first to provide the issues
        """
        #TODO: check for self.issues, report sensible error if check_all hasn't
        # been run yet.

        if not self.issues:
            print("Style check passed with flying colors!")
            return

        num_issues = sum(len(p[1]) for p in self.issues)

        print("Found %d issues:" % num_issues)
        for pagename,issuelist in self.issues:
            print("Page: %s" % pagename)
            c = collections.Counter(issuelist)
            for i, count_i in c.items():
                if i[0]=="Unplain Phrase":
                    print("   Discouraged phrase: %s (%d instances); suggest '%s' instead." %
                                    ( i[1], count_i, self.disallowed_phrases[i[1].lower()] ))
                elif i[0]=="Unplain Word":
                    print("   Discouraged word: %s (%d instances); suggest '%s' instead." %
                                    ( i[1], count_i, self.disallowed_words[i[1].lower()] ))
                else:
                    print("   %s: %s (%d instances)" % (i[0], i[1], count_i))


    def get_overrides(self, soup):
        """
        Look for overrides in the text to make exceptions for specific style
        rules. Returns a set of rule strings to ignore for this block.
        """

        overrides = set()
        comments = soup.find_all(string=lambda text:isinstance(text,Comment))
        for comment in comments:
            m = re.match(OVERRIDE_COMMENT_REGEX, comment)
            if m:
                new_overrides = m.group(1).split(",")
                new_overrides = {o.strip() for o in new_overrides}
                logger.info("Overrides found: %s" % new_overrides)
                overrides |= new_overrides

        return overrides

    def check_passage(self, passage, overrides):
        """Checks an individual string of text for style issues."""
        issues = []
        logging.debug("Checking passage %s" % passage)
        #tokens = nltk.word_tokenize(passage)
        tokens = self.tokenize(passage)
        for t in tokens:
            if t.lower() in self.disallowed_words:
                if t.lower() in overrides:
                    logger.info("Unplain word violation %s overridden" % t)
                    continue
                issues.append( ("Unplain Word", t.lower()) )

        for phrase,sub in self.disallowed_phrases.items():
            if phrase.lower() in self.depunctuate(passage):
                if phrase.lower() in overrides:
                    logger.info("Unplain phrase violation %s overridden" % t)
                    continue
                issues.append( ("Unplain Phrase", phrase.lower()) )

        return issues


def main(cli_args, config):
    target = DactylTarget(config, name=cli_args.target)
    checker = DactylStyleChecker(target, config)

    issues = checker.check_all()
    checker.report()
    if issues:
        exit(1)
    else:
        exit(0)


def dispatch_main():
    cli = DactylCLIParser(DactylCLIParser.UTIL_STYLE)
    config = DactylConfig(cli.cli_args)
    main(cli.cli_args, config)


if __name__ == "__main__":
    dispatch_main()



################################################################################
## Add Markdown Class to Divs filter                                          ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Finds raw divs in the markdown and adds the markdown=1 attribute to them   ##
## so that HTML inside those divs gets parsed as markdown.                    ##
## Some flavors of markdown do this automatically, so this provides           ##
## compatibility with those.                                                  ##
################################################################################

def filter_markdown(md, **kwargs):
    """Python markdown requires markdown="1" on HTML block elements
       that contain markdown. AND there's a bug where if you use
       markdown.extensions.extra, it replaces code fences in HTML
       block elements with garbled text."""
    def add_markdown_class(m):
        if m.group(0).find("markdown=") == -1:
            return m.group(1) + ' markdown="1">'
        else:
            return m.group(0)

    logger.info("... adding markdown class to embedded divs...")
    md = re.sub(r"(<div[^>]*)>", add_markdown_class, md)
    return md



################################################################################
## Multicode Tabs 2 filter                                                    ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Finds multicode tab sections and turns them into properly-formatted        ##
## HTML syntax to use with minitabs jQuery                                    ##
################################################################################
import re
import logging

MC_START_REGEX = re.compile(r"<!--\s*MULTICODE_BLOCK_START\s*-->")
MC_END_REGEX = re.compile(r"<!--\s*MULTICODE_BLOCK_END\s*-->")

def filter_html(html, mode="html", **kwargs):
    """
    Turn multicode comments into a div (after markdown inside is parsed). You
    can use this div for styling even in PDF format. Doesn't apply to Markdown
    since most parsers won't parse markdown inside HTML blocks.
    """

    if mode == "md":
        return html

    html = re.sub(MC_START_REGEX, "<div class='multicode'>", html)
    html = re.sub(MC_END_REGEX, "</div>", html)
    return html

def filter_soup(soup, mode="html", **kwargs):
    """Turn a multicode block into the correct syntax for minitabs, but only
       in the HTML version."""
    if mode != "html":
        return

    multicodes = soup.find_all(class_="multicode")
    index1 = 0
    for cb_area in multicodes:
        cb_area["id"] = "code-%d" % index1

        codetabs_ul = soup.new_tag("ul")
        codetabs_ul["class"] = "codetabs"
        cb_area.insert(0,codetabs_ul)

        pres = cb_area.find_all("pre")
        index2 = 0
        for pre in pres:
            #make a unique ID for this code sample
            linkid = "code-%d-%d" % (index1, index2)

            #wrap this code sample in an ID'd div
            code_sample_wrapper = soup.new_tag("div", id=linkid)
            code_sample_wrapper["class"] = "code_sample"
            code_sample_wrapper["style"] = "position: static;"
            pre.wrap(code_sample_wrapper)

            #add a link to the tabs ul
            linkback = soup.new_tag("a", href=("#%s" % linkid))
            linkback_li = soup.new_tag("li")
            linkback_li.append(linkback)
            codetabs_ul.append(linkback_li)

            #find the text label for this sample
            prev_p = code_sample_wrapper.find_previous_sibling("p")
            try:
                label = "".join(prev_p.em.strings)
            except AttributeError:
                label = "Code Sample %d-%d" % (index1, index2)
            linkback.string = label
            prev_p.decompose()

            index2 += 1

        index1 += 1



################################################################################
## Substitute Links filter                                                    ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2017                                          ##
##                                                                            ##
## Replaces the link substitution feature built into Dactyl < 0.4.0 with a    ##
## filter to do about the same.                                               ##
################################################################################

import re

LINK_SUBS_FIELD = "link_subs"
LINK_RE_SUBS_FIELD = "link_re_subs"
PARAMETER_REPLACE_FIELD = "replace_parameter_links"
IMAGE_SUBS_FIELD = "image_subs"
IMAGE_RE_SUBS_FIELD = "image_re_subs"
IMAGE_LINK_REGEX = re.compile(r"^[^.]+\.(png|jpg|jpeg|gif|svg)", re.I)

def filter_soup(soup, currentpage={}, target={}, pages=[], logger=None, **kwargs):
    """
    Replaces links and image hrefs in the current page, based on a substitution
    map in the target or page settings. Also looks into values in the current
    page's metadata and replaces links there, in case the template uses fields
    from the current page's metadata key/values.
    """
    globals()["logger"] = logger
    # currentpage already includes link subs inherited from the target
    if LINK_SUBS_FIELD in currentpage:
        link_subs = currentpage[LINK_SUBS_FIELD]

        if (PARAMETER_REPLACE_FIELD in currentpage and
                currentpage[PARAMETER_REPLACE_FIELD]):
            substitute_parameter_links(currentpage, link_subs)
        substitute_links(soup, link_subs)

    if LINK_RE_SUBS_FIELD in currentpage:
        link_re_subs = currentpage[LINK_RE_SUBS_FIELD]
        re_sub_links(soup, link_re_subs)

    if IMAGE_SUBS_FIELD in currentpage:
        image_subs = currentpage[IMAGE_SUBS_FIELD]
        substitute_images(soup, image_subs)
        substitute_links(soup, image_subs)

    if IMAGE_RE_SUBS_FIELD in currentpage:
        image_re_subs = currentpage[IMAGE_RE_SUBS_FIELD]
        re_sub_images(soup, image_re_subs)
        re_sub_links(soup, image_re_subs)

def substitute_links(soup, link_subs):
    """
    Takes a map of needle:replacement strings and changes the href values of
    <a> tags in the soup, so that any that start with the needle are changed to
    start with the replacement instead (preserving the remainder).
    """
    links = soup.find_all("a", href=True)
    for link in links:
        for needle, replacement in link_subs.items():
            if link["href"][:len(needle)] == needle:
                new_href = replacement + link["href"][len(needle):]
                logger.info("... replacing link '%s' with '%s'" %
                            (link["href"], new_href) )
                link["href"] = new_href

def substitute_images(soup, image_subs):
    """
    Takes a map of needle:replacement strings and changes the src of <img>
    tags in the soup so that any that match the needle are changed to use the
    replacement instead.
    """
    images = soup.find_all("img")
    for img in images:
        for needle, replacement in image_subs.items():
            if needle == img["src"]:
                logger.info("... replacing image '%s' with '%s'" %
                            (img["src"], replacement) )
                img["src"] = replacement


def re_sub_images(soup, image_re_subs):
    """
    Takes a map of regular expressions to regular-expression replacements and
    changes the src of any <img> tags in the soup by doing regular-expression
    match/replace.
    """
    images = soup.find_all("img", src=True)
    for img in images:
        for regex,replace_pattern in image_re_subs.items():
            m = re.match(regex, img["src"])
            if m:
                new_path = re.sub(regex, replace_pattern, img["src"])
                logger.info("... replacing image '%s' with '%s'" %
                            (img["src"], new_path) )
                img["src"] = new_path

def re_sub_links(soup, link_re_subs):
    """
    Takes a map of regular expressions to regular-expression replacements and
    changes the href of any <a> tags in the soup by doing regular-expression
    match/replace.
    """
    links = soup.find_all("a", href=True)
    for link in links:
        for regex,replace_pattern in link_re_subs.items():
            m = re.match(regex, link["href"])
            if m:
                new_path = re.sub(regex, replace_pattern, link["href"])
                logger.info("... replacing link '%s' with '%s'" %
                            (link["href"], new_path) )
                link["href"] = new_path

RESERVED_PAGE_KEYS = [
    "html",
    "md",
    "category",
    "targets"
]
def substitute_parameter_links(currentpage, link_subs):
    """
    Takes a map of needle:replacement link substitutions and applies them to
    string values in the current page's metadata parameters.
    """
    for field,val in currentpage.items():
        if field in RESERVED_PAGE_KEYS:
            continue;
        if type(val) != str:
            continue
        for needle, replacement in link_subs:
            if val[:len(needle)] == needle:
                new_val = replacement + val[len(needle):]
                logger.info(("... replacing field '%s'; replacing value "+
                            "'%s' with '%s'") %  (field, val, new_val) )
                currentpage[field] = new_val

class MDLink: # Used for link substitution on markdown
    """A markdown link, in one of the following formats:
    - [label](url "title") — an inline link (title optional)
    - ![alt text](url) — an inline image link
    - [ref]: url — a reference link definition (could be an image or not)


    Affects, but does not include, reference link instances, such as:
    - [ref][]
    - [label][ref]
    """

    MD_LINK_REGEX = re.compile(
        r"(\[([^\]]+)\]\(([^:)]+)\)|\[([^\]]+)\]:\s*(\S+)$)", re.MULTILINE)

    MD_IMG_REGEX = re.compile(
        r"!(\[([^\]]+)\]\(([^:)]+)\)|\[([^\]]+)\]:\s*(\S+)$)", re.MULTILINE)

    def __init__(self, fullmatch, label, url, label2, url2):
        self.fullmatch = fullmatch
        if fullmatch[:1] == "!":
            self.is_img = True
        else:
            self.is_img = False
        if label:
            self.label = label
            self.url = url
            self.is_reflink = False
        elif label2:
            self.label = label2
            self.url = url2
            self.is_reflink = True

    def to_markdown(self):
        """Re-represent self as a link in markdown syntax"""
        s = "[" + self.label + "]"
        if self.is_reflink:
            s += ": " + self.url
        else:
            s += "(" + self.url + ")"
        return s

def filter_markdown(md, mode="html", currentpage={}, logger=None, **kwargs):
    """
    In "Githubify" mode, we need to do link substitution on the Markdown itself
    since we won't have an opportunity to do it on the HTML output.
    """
    globals()["logger"] = logger
    if mode != "md":
        return md

    if LINK_SUBS_FIELD in currentpage:
        link_subs = currentpage[LINK_SUBS_FIELD]
        md = substitute_md_links(md, link_subs)

    if IMAGE_SUBS_FIELD in currentpage:
        image_subs = currentpage[IMAGE_SUBS_FIELD]
        md = substitute_md_images(currentpage, image_subs)
        md = substitute_md_links(currentpage, image_subs)

    if LINK_RE_SUBS_FIELD in currentpage:
        link_re_subs = currentpage[LINK_RE_SUBS_FIELD]
        md = substitute_md_links(md, link_re_subs, regex_search=True)

    if IMAGE_RE_SUBS_FIELD in currentpage:
        image_re_subs = currentpage[IMAGE_RE_SUBS_FIELD]
        md = substitute_md_images(md, image_re_subs, regex_search=True)
        md = substitute_md_links(md, link_re_subs, regex_search=True)

    return md

def substitute_md_links(md, link_subs, do_images=False, regex_search=False):
    if do_images:
        regex = MDLink.MD_IMG_REGEX
    else:
        regex = MDLink.MD_LINK_REGEX
    links = [MDLink(*m) for m in regex.findall(md)]

    for link in links:
        for needle, replacement in link_subs.items():
            if regex_search:
                m = re.match(needle, link.url)
                if m:
                    new_path = re.sub(needle, replacement, link.url)
                    link.url = new_path
                    logger.info("... replacing link '%s' with '%s'" %
                                (link.fullmatch, link.to_markdown()) )
                    md = md.replace(link.fullmatch, link.to_markdown() )
            else:
                if link.url[:len(needle)] == needle:
                    link.url = replacement + link.url[len(needle):]
                    md = md.replace(link.fullmatch, link.to_markdown())

    return md

def substitute_md_images(md, image_subs, regex_search=False):
    return substitute_md_links(md, image_subs, do_images=True, regex_search=regex_search)



################################################################################
## Dactyl Target Class
##
## Defines a group of pages that are built together.
################################################################################

from dactyl.common import *
from dactyl.config import DactylConfig
from dactyl.openapi import ApiDef
from dactyl.page import DactylPage

class DactylTarget:
    def __init__(self, config, name=None, inpages=None, spec_path=None):
        assert isinstance(config, DactylConfig)
        self.config = config
        self.pages = None
        self.cover = None

        # Set self.data based on the input type
        if inpages is not None:
            self.from_adhoc(inpages)
        elif spec_path is not None:
            self.from_openapi(spec_path)
        else:
            self.from_config(name)
        self.name = self.data["name"]

        # Load *ALL* pages in config, in case some of them need to be referenced
        # by pages in this target
        self.config.load_pages()


    def from_config(self, name=None):
        """
        Pick a target from the existing config file, by its short name.
        If no name is provided, use whichever target is listed first.
        """
        if name is None:
            if len(self.config["targets"]) == 0:
                logger.critical("No targets found. Either specify a config file or --pages")
                exit(1)
            self.data = self.config["targets"][0]
        else:
            try:
                self.data = next(t for t in self.config["targets"] if t["name"] == name)
            except StopIteration:
                logger.critical("Unknown target: %s" % name)
                exit(1)

    def from_adhoc(self, inpages):
        """
        Create an ad-hoc target from a list of input pages, generally passed as
        a list of files from the commandline.
        """
        t = {
            "name": ADHOC_TARGET,
            "display_name": UNTITLED_TARGET,
        }

        # if len(inpages) == 1:
        #     t["display_name"] = guess_title_from_md_file(inpages[0])

        for inpage in inpages:
            # Figure out the actual filename and location of this infile
            # and set the content source dir appropriately
            in_dir, in_file = os.path.split(inpage)
            self.config["content_path"] = in_dir

            # Figure out what html file to output to
            ENDS_IN_MD = re.compile("\.md$", re.I)
            if re.search(ENDS_IN_MD, in_file):
                out_html_file = re.sub(ENDS_IN_MD, ".html", in_file)
            else:
                out_html_file = in_file+".html"

            new_page = {
                "md": in_file,
                "html": out_html_file,
                "targets": [ADHOC_TARGET],
                "category": "Pages",
                "pp_dir": in_dir,
            }
            self.config["pages"].append(new_page)

        self.data = t
        self.config["targets"].append(t)

    def from_openapi(self, spec_path):
        """
        Create a target from an API spec path.
        """

        openapi = ApiDef.from_path(spec_path)
        t = {
            "name": openapi.api_slug,
            "display_name": openapi.api_title,
        }

        self.config["pages"] = [{
            "openapi_specification": spec_path,
            "api_slug": openapi.api_slug,
            "targets": [openapi.api_slug],
        }]
        self.data = t
        openapi.add_metadata(self.data)
        self.config["targets"].append(t)

    def slug_name(self, fields_to_use=[], separator="-"):
        """Make a name for the target that's safe for use in URLs & filenames,
        from human-readable fields"""
        filename_segments = []
        for fieldname in fields_to_use:
            if fieldname in self.data.keys():
                filename_segments.append(slugify(self.data[fieldname]))

        if filename_segments:
            return separator.join(filename_segments)
        else:
            return slugify(self.name)

    def es_index_name(self):
        return self.slug_name(self.config["es_index_fields"],
                              self.config["es_index_separator"])

    def add_cover(self):
        """
        Instantiate the default cover page so self.load_pages() can include it.
        """
        coverpage_data = self.config["cover_page"]
        coverpage_data["targets"] = [self.name]
        skip_pp = self.config.get("skip_preprocessor", False)
        more_filters = self.data.get("filters", [])
        self.cover = DactylPage(self.config, coverpage_data, skip_pp, more_filters)

    def default_pdf_name(self):
        """Choose a reasonable name for a PDF file in case one isn't specified."""
        return self.slug_name(self.config["pdf_filename_fields"],
                              self.config["pdf_filename_separator"]) + ".pdf"


    def should_include(self, page):
        """Report whether a given page should be part of this target"""
        if "targets" not in page:
            return False
        if self.name in page["targets"]:
            return True
        else:
            return False

    def gain_fields(self, fields):
        """
        Add the provided fields to this target's definition
        """
        merge_dicts(fields, self.data, RESERVED_KEYS_TARGET)
        if "display_name" in fields: # Exception to reserved key rule
            self.data["display_name"] = fields["display_name"]


    def expand_openapi_spec(self, page_data):
        """Expand OpenAPI Spec placeholders into a full page list"""
        assert OPENAPI_SPEC_KEY in page_data.keys()

        logger.debug("Expanding OpenAPI spec from placeholder: %s"%page_data)
        api_slug = page_data.get(API_SLUG_KEY, None)
        extra_fields = {}
        merge_dicts(self.data, extra_fields, RESERVED_KEYS_TARGET)
        merge_dicts(page_data, extra_fields, [OPENAPI_SPEC_KEY, API_SLUG_KEY])

        template_path = page_data.get(OPENAPI_TEMPLATE_PATH_KEY, None)
        swagger = ApiDef.from_path(page_data[OPENAPI_SPEC_KEY], api_slug,
                                       extra_fields, template_path)
        skip_pp = self.config.get("skip_preprocessor", False)
        made_pages = []
        for p in swagger.create_pagelist():
            more_filters = self.data.get("filters", [])
            po = DactylPage(self.config, p, skip_pp, more_filters)
            # Special case for filters; concatenate filter lists,
            # with target's filters first.
            if "filters" in self.data:
                po.gain_filters(self.data["filters"])
            made_pages.append(po)
        return made_pages

    def load_pages(self):
        """
        Find the set of pages that this target should include.  At this time,
        we expand OpenAPI spec placeholders into individual pages, read
        frontmatter, and so on. Save it to self.pages for later reference.
        """
        pages = []
        if self.cover:
            merge_dicts(self.data, self.cover.data, RESERVED_KEYS_TARGET)
            pages.append(self.cover)
        for i, page_candidate in enumerate(self.config.page_cache):
            if page_candidate == OPENAPI_SPEC_PLACEHOLDER:
                page_data = self.config["pages"][i]
                logger.debug("loading API spec: {pg}".format(pg=page_data))
            elif page_candidate == NOT_LOADED_PLACEHOLDER:
                page_data = self.config["pages"][i]
                # A not-loaded page is fine as long as this target doesn't
                # need it.
            else:
                page_data = page_candidate.data

            if not self.should_include(page_data):
                continue # Skip out-of-target pages;
            if page_candidate == NOT_LOADED_PLACEHOLDER:
                recoverable_error("Requires a page that failed to load: %s" %
                                  page_data, self.config.bypass_errors)
                continue # If we bypass errors, just omit the unloaded page

            # Expand OpenAPI Spec placeholders into full page lists
            if OPENAPI_SPEC_KEY in page_data.keys():
                try:
                    swagger_pages = self.expand_openapi_spec(page_data)
                    logger.debug("Adding generated OpenAPI reference pages: %s" %
                                    [p.data["name"] for p in swagger_pages])
                    pages += swagger_pages
                except Exception as e:
                    recoverable_error("Error when parsing OpenAPI definition %s: %s" %
                                      (page_data, e), self.config.bypass_errors)
                    # Omit the API def from the page list if an error occurred

            # Load cached page object; pass on target key-values,
            # then add it to the list
            else:
                merge_dicts(self.data, page_data, RESERVED_KEYS_TARGET)
                # Special case for filters; concatenate filter lists,
                # with target's filters first.
                if "filters" in self.data:
                    page_candidate.gain_filters(self.data["filters"])
                pages.append(page_candidate)

        # Check for pages that would overwrite each other
        html_outs = set()
        for p in pages:
            if "html" not in p.data.keys():
                logger.error("page has no html somehow? %s"%p.data)
            if p.data["html"] in html_outs:
                recoverable_error(("Repeated output filename '%s'. The earlier "+
                    "instances will be overwritten") % p.data["html"],
                    self.config.bypass_errors)
            html_outs.add(p.data["html"])

        self.pages = pages

        # Special case for ad-hoc targets with one page: rename to page name
        # if no title was provided.
        if len(self.pages) == 1 and self.name == UNTITLED_TARGET:
            self.name = self.pages[0].data["name"]
            self.data["display_name"] = self.name

        self.find_hierarchy()
        return pages

    def find_hierarchy(self):
        """
        Adds "children" arrays to pages to mirror "parent" links.
        """
        for p in self.pages:
            if "parent" in p.data.keys():
                try:
                    parent = next((pg for pg in self.pages if pg.data["html"] == p.data["parent"]))
                except StopIteration:
                    logger.warning("parent '%s' value not found in this target"%p.data["parent"])
                    continue

                if "children" in parent.data.keys():
                    if type(parent.data["children"]) != list:
                        logger.warning("hierarchy: page '%s' has explicit children field, not modifying."%parent.data["html"])
                        continue
                    else:
                        # Add this child to the parent's existing list if it's
                        # not already there.
                        if p.data not in parent.data["children"]:
                            parent.data["children"].append(p.data)
                else:
                    # Start a new child list at the parent
                    parent.data["children"] = [p.data]

            if "children" not in p.data.keys():
                # Start an empty list of children
                p.data["children"] = []

            p.data["is_ancestor_of"] = self.make_ancestor_lookup(p)

    @staticmethod
    def make_ancestor_lookup(p):
        # logger.debug("defining is_ancestor_of for %s"%p)
        def is_ancestor_of(html):
            if "children" not in p.data.keys():
                return False
            if type(p.data["children"]) != list:
                return False
            for kid in p.data["children"]:
                if kid["html"] == html:
                    return True
                if kid["is_ancestor_of"](html):
                    return True
            return False
        return is_ancestor_of

    def categories(self):
        """Produce an ordered, de-duplicated list of categories from
           this target's page list"""
        categories = []
        for page in self.pages:
            if "category" in page.data and page.data["category"] not in categories:
                categories.append(page.data["category"])
        logger.debug("categories: %s" % categories)
        return categories



from collections import defaultdict

import requests
from jinja2 import BaseLoader, FileSystemLoader, TemplateNotFound, Template
from urllib.parse import urlparse
from dactyl.common import *

class FrontMatterRemoteLoader(BaseLoader):
    def __init__(self):
        self.baseurl = None
        self.fm_map = defaultdict(dict) # Save frontmatter here, if any
        self.twolines = {} # Save first two lines so we can try guessing titles

    def get_source(self, environment, template):
        """Fetch a remote markdown file and return its contents"""
        parsed_url = urlparse(template)

        # save this base URL if it's new; that way we can try to let remote
        # templates inherit from other templates at related URLs
        if parsed_url.scheme or (self.baseurl is None):
            base_path = os.path.dirname(parsed_url.path)
            self.baseurl = (
                parsed_url[0],
                parsed_url[1],
                base_path,
                parsed_url[3],
                parsed_url[4],
                parsed_url[5],
            )
        # if we're at read_markdown_remote without a scheme, it's probably a remote
        # template trying to import another template, so let's assume it's from the
        # base path of the imported template
        if not parsed_url.scheme:
            url = os.path.join(self.baseurl, template)
        else:
            url = template

        response = requests.get(url)
        if response.status_code == 200:
            text, frontmatter = parse_frontmatter(response.text)
            self.fm_map[template] = frontmatter
            self.twolines[template] = text.split("\n", 2)[:2]
            return text, url, None
        else:
            raise TemplateNotFound(template)

class FrontMatterFSLoader(FileSystemLoader):
    def __init__(self, searchpath, encoding='utf-8', followlinks=False):
        super().__init__(searchpath, encoding, followlinks)
        self.fm_map = defaultdict(dict)
        self.twolines = {}

    def get_source(self, environment, template):
        text, filename, uptodate = super().get_source(environment, template)
        text, frontmatter = parse_frontmatter(text)
        self.fm_map[template] = frontmatter
        self.twolines[template] = text.split("\n", 2)[:2]
        return text, filename, uptodate



################################################################################
## Demote Headers filter                                                      ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2017                                          ##
##                                                                            ##
## Demote all headers one level. This can be useful to make PDF docs render   ##
## a nice link hierarchy if your templates provide category headers at h1 but ##
## your docs individually start at the h1 level.                              ##
################################################################################

DEMOTE_FIELD = "demote_headers_pdf_only"

def filter_html(html, mode="html", target={}, **kwargs):
    if (mode == "html" and
        DEMOTE_FIELD in target and
        target[DEMOTE_FIELD]):
        # Don't bother then
        return html

    html = html.replace("<h5", "<h6")
    html = html.replace("<h4", "<h5")
    html = html.replace("<h3", "<h4")
    html = html.replace("<h2", "<h3")
    html = html.replace("<h1", "<h2")
    return html



################################################################################
## Badges filter                                                              ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2016                                          ##
##                                                                            ##
## Looks for links with the title text "BADGE" and makes them into badges.    ##
## The alt text must be in the form of <badgelefthalf>:<badgerighthalf> and   ##
## the left half can't contain a colon.                                       ##
################################################################################
import re
import logging
from urllib.parse import quote as urlescape

BADGE_REGEX = re.compile("BADGE_(BRIGHTGREEN|GREEN|YELLOWGREEN|YELLOW|ORANGE|RED|LIGHTGREY|BLUE|[0-9A-Fa-f]{6})")

def filter_soup(soup, **kwargs):
    """replace underscores with dashes in h1,h2,etc. for backwards compatibility"""

    badge_links = soup.find_all(name="a", title=BADGE_REGEX)

    for b in badge_links:
        badge_label = b.string
        if not badge_label:
            badge_label = "".join(b.strings)
        if not badge_label:
            logging.warning("Badge link with no string: %s" % b)
            continue
        if ":" not in badge_label:
            logging.warning("Badge link specified with no ':' in link: %s" % b.string)
            continue

        badge_color = BADGE_REGEX.match(b["title"]).group(1).lower()
        badge_left, badge_right = [urlescape(s.strip()).replace("-","--")
                                   for s in badge_label.split(":", 1)]
        badge_url = "https://img.shields.io/badge/%s-%s-%s.svg" % (
                    badge_left, badge_right, badge_color)

        img = soup.new_tag("img", src=badge_url, alt=badge_label)
        img["class"]="dactyl_badge"
        b.clear()
        b.append(img)
        b["title"] = badge_label
        if not b["href"]:
            del b["href"]



################################################################################
# Dactyl common code
#
# Imports and utilities shared across multiple pieces of Dactyl
################################################################################

# The ElasticSearch templates need to write *actual* JSON and not YAML
import json
import logging
import os
import re
import time
import traceback

from pkg_resources import resource_stream

import ruamel.yaml
yaml = ruamel.yaml.YAML(typ="safe")

import gettext

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.propagate = False

PDF_USE_DEFAULT = "__DEFAULT_FILENAME__"
NO_PDF = "__NO_PDF__"
ES_USE_DEFAULT = "__ES_USE_DEFAULT__"
NO_ES_UP = False

# These fields are special, and pages don't inherit them directly
RESERVED_KEYS_TARGET = [
    "name",
    "display_name",
    "pages",
]
ADHOC_TARGET = "__ADHOC__"
UNTITLED_TARGET = "(Untitled)"

PROVIDED_FILENAME_KEY = "__dactyl_provided_filename__"
ES_EVAL_KEY = "__dactyl_eval__"
OPENAPI_SPEC_KEY = "openapi_specification"
OPENAPI_TEMPLATE_PATH_KEY = "openapi_md_template_path"
OPENAPI_SPEC_PLACEHOLDER = "__OPENAPI_SPEC_PLACEHOLDER__"
NOT_LOADED_PLACEHOLDER = "__NOT_LOADED_PLACEHOLDER__"
API_SLUG_KEY = "api_slug"
BUILTIN_ES_TEMPLATE = "templates/template-es.json"

DEFAULT_SERVER_PORT = 32289 # "DACTY" in T-9


def recoverable_error(msg, bypass_errors, error=None):
    """Logs a warning/error message and exits if bypass_errors==False"""
    if not bypass_errors and error is not None:
        traceback.print_tb(error.__traceback__)
    logger.error(msg)
    if not bypass_errors:
        exit(1)

# Note: this regex means non-ascii characters get stripped from filenames,
#  which is not preferable when making non-English filenames.
unacceptable_chars = re.compile(r"[^A-Za-z0-9._ ]+")
whitespace_regex = re.compile(r"\s+")
def slugify(s):
    s = re.sub(unacceptable_chars, "", s)
    s = re.sub(whitespace_regex, "_", s)
    if not s:
        s = "_"
    return s


def parse_frontmatter(text):
    """Separate YAML frontmatter, if any, from a string, and return the
    text separate from the parsed front-matter."""
    if len(text) < 6:
        logger.debug("...too short for frontmatter")
        return text, {}

    if text[:3] == "---" and text.find("---", 3) != -1:
        logger.debug("...has front matter")
        raw_frontmatter = text[3:text.find("---", 3)]
        frontmatter = yaml.load(raw_frontmatter)
        # Map some Jekyll-specific frontmatter variables to their Dactyl equivs
        if "title" in frontmatter.keys():
            # We don't care about the Jekyll "page.name" field so it's OK to
            #   overwrite it.
            frontmatter["name"] = frontmatter["title"]
        if "categories" in frontmatter.keys() and len(frontmatter["categories"]):
            frontmatter["category"] = frontmatter["categories"][0]
        logger.debug("Loaded frontmatter: %s"% frontmatter)

        return text[text.find("---", 3)+4:], frontmatter
    else:
        logger.debug("...no front matter detected")
        return text, {}

def merge_dicts(default_d, specific_d, reserved_keys_top=[]):
    """
    Extend specific_d with values from default_d (recursively), keeping values
    from specific_d where they both exist. (This is like dict.update() but
    without overwriting duplicate keys in the updated dict.)

    reserved_keys_top is only used at the top level, not recursively
    """
    for key,val in default_d.items():
        if key in reserved_keys_top:
            continue
        if key not in specific_d.keys():
            specific_d[key] = val
        elif type(specific_d[key]) == dict and type(val) == dict:
                merge_dicts(val, specific_d[key])
        #else leave the key in the specific_d



#!/usr/bin/env python3

"""setuptools based setup module"""

from setuptools import setup

long_description = open('README.md', encoding="utf-8").read()

with open("dactyl/version.py", encoding="utf-8") as versionfile:
    exec(versionfile.read())

setup(
    name='dactyl',
    version=__version__,
    description='Tools to generate documentation.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ripple/dactyl',
    author='Ripple',
    author_email='rome@ripple.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
        'Topic :: Documentation',
    ],
    keywords='documentation',
    packages=[
        'dactyl',
    ],
    entry_points={
        'console_scripts': [
            'dactyl_build = dactyl.dactyl_build:dispatch_main',
            'dactyl_link_checker = dactyl.dactyl_link_checker:dispatch_main',
            'dactyl_style_checker = dactyl.dactyl_style_checker:dispatch_main',
        ],
    },
    install_requires=[
        'argparse',
        'beautifulsoup4',
        'jinja2>=2.11',
        'Markdown>=3.0.1',
        'ruamel.yaml',
        'requests',
        'watchdog'
    ],
    package_data={
        '': ["templates/*", "default-config.yml"],
    }
)



################################################################################
## XRefs: Intelligent Crossreferences filter                                  ##
## Author: Rome Reginelli                                                     ##
## Copyright: Ripple Labs, Inc. 2017                                          ##
##                                                                            ##
## Looks for syntax matching the following format:                            ##
##     [optional text](XREF: some-link.html#fragment)                         ##
## and interprets it as cross-references. If some-link.html is a file in the  ##
## current target it becomes a normal hyperlink. If the link text is [] (that ##
## is, blank) it gets replaced with the title of the page.                    ##
## (Note: we can't look up section titles as that would require parsing the   ##
## cross-referenced page and could lead to an infinite recursion loop if two  ##
## pages cross-ferenced each other.)                                          ##
##   If the file isn't part of the current target but is part of another      ##
## target, it becomes a non-hyperlink cross reference to the page in the      ##
## first target that DOES have it. For example:                               ##
##      "Some Link Title" in _A Target Containing Some Link_                  ##
################################################################################
import re
from logging import warning

# match anything starting with XREF:/xref:, split by the # if there is one
# dropping any excess whitespace
xref_regex = re.compile(r"^\s*xref:\s*(?P<xref_file>[^#]+)(?P<xref_frag>#\S+)?\s*?$", re.I)

class BadXref(Exception):
    pass

def find_file_in_target(fname, targetname, config):
    if fname[-3:] == ".md":
        # look by markdown file first
        for page in config["pages"]:
            if "md" not in page:
                continue
            elif ("/" in fname and page["md"] == fname # try to match md file by exact path
                    and targetname in page.get("targets",[]) # the page appears in this target
                    and page.get("html","") ): # and finally, the page has an html filename
                return page
            elif ( page["md"].split("/")[-1] == fname # match md filename in any directory
                    and targetname in page.get("targets",[])
                    and page.get("html","") ):
                return page

    for page in config["pages"]:
        if "html" not in page:
            continue
        elif page["html"] != fname:
            continue
        if targetname in page.get("targets", []):
            return page
    else:
        return False

def find_file_in_any_target(fname, config):
    if fname[-3:] == ".md":
        #print("finding in any target by md")
        # look by markdown file first
        for page in config["pages"]:
            if "md" not in page:
                continue
            elif ("/" in fname and page["md"] == fname # try to match md file by exact path
                   and page.get("targets",[]) # page must appear in some target
                   and page.get("html","") ): # and page must have an html filename
                return page
            elif ( page["md"].split("/")[-1] == fname # match md filename in any folder
                   and page.get("targets",[])
                   and page.get("html","") ):
                return page

    # look by HTML file if it didn't end in .md or if we didn't find it yet
    for page in config["pages"]:
        if "html" not in page:
            continue
        elif page["html"] == fname and page.get("targets", []):
            #page has to have "some" target(s) for it to be worthwhile
            return page
    else:
        return False

def lookup_display_name(targetname, config):
    for t in config["targets"]:
        if "name" in t and t["name"] == targetname:
            display_name = "%s %s %s %s %s" % (
                t.get("display_name", ""),
                t.get("product", ""),
                t.get("version", ""),
                t.get("guide", ""),
                t.get("subtitle", "")
            )
            if display_name.strip():
                return display_name
            else:
                warning("Target has no display_name/product/version/guide: %s" % targetname)
                return targetname
    else:
        warning("Target not found: %s" % targetname)
        return targetname

def filter_soup(soup, target={"name":""}, currentpage={},
        config={"pages":[]}, logger=None, **kwargs):
    """Look for cross-references and replace them with not-hyperlinks if they
       don't exist in the current target."""

    prefix = target.get("prefix", "")
    xrefs = soup.find_all(href=xref_regex)

    for xref in xrefs:
        m = xref_regex.match(xref.attrs["href"])
        xref_file = m.group("xref_file")
        xref_frag = m.group("xref_frag") or ""

        xref_page = find_file_in_target(xref_file, target["name"], config)
        if xref_page == False:
            # Cross-referenced page isn't part of this target
            xref_page = find_file_in_any_target(xref_file, config)
            if not xref_page:
                if currentpage and "md" in currentpage:
                    page_name = currentpage["md"]
                else:
                    page_name = currentpage
                raise BadXref(("xref to missing file: '%s' in page '%s'. "+
                "Maybe it's not in the Dactyl config file?")%(xref_file, page_name))
            xref_target_shortname = xref_page["targets"][0]

            ref_target = lookup_display_name(xref_target_shortname, config)

            link_label = " ".join([s for s in xref.stripped_strings])
            # TODO: consider preserving non-string contents in the link label
            # If a link label wasn't provided, generate one from the page name
            if not link_label.strip():
                link_label = xref_page.get("name", None)
                if link_label is None:
                    if currentpage and "md" in currentpage:
                        page_name = currentpage["md"]
                    else:
                        page_name = currentpage
                    if config.bypass_errors:
                        logger.warning(("Xref without label from page "+
                        "'%s' to '%s'. Either provide link text or make sure "+
                        "the target page's name is defined.")%(page_name,m.group(0)))
                        link_label = xref_page.get("html","(Unknown)")
                    else:
                        raise BadXref(("Couldn't get label for xref from page "+
                        "'%s' to '%s'. Either provide link text or make sure "+
                        "the target page's name is defined.")%(page_name,m.group(0)))
            link_label = link_label.strip()

            # "Link Label" in _Target Display Name_
            span = soup.new_tag("span")
            span.attrs["class"] = "dactyl_xref"
            span.string = '"%s" in the ' % link_label
            em = soup.new_tag("em")
            em.string = ref_target
            span.append(em)
            xref.replace_with(span)

        else:
            # The xref is on-target
            # First fix the hyperlink. Use the HTML (in case of link-by-md):
            xref.attrs["href"] = prefix+xref_page["html"]+xref_frag
            # If this link's label is only whitespace, fix it
            xref_contents = "".join([str(c) for c in xref.contents]).strip()
            # logger.debug("xref contents: '%s'"%xref_contents)
            if not xref_contents:
                xref_label = xref_page.get("name", "")
                xref.string = xref_label



HTTP_METHODS = [
    "get",
    "put",
    "post",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
]

HTTP_METHODS_WITH_REQ_BODIES = [
    "put",
    "post",
    "delete",
    "patch",
]

HTTP_STATUS_CODES = {
    "100": "Continue",
    "101": "Switching Protocols",
    "102": "Processing",
    "103": "Early Hints",
    "200": "OK",
    "201": "Created",
    "202": "Accepted",
    "203": "Non-Authoritative Information",
    "204": "No Content",
    "205": "Reset Content",
    "206": "Partial Content",
    "207": "Multi-Status",
    "208": "Already Reported",
    "226": "IM Used",
    "300": "Multiple Choices",
    "301": "Moved Permanently",
    "302": "Found",
    "303": "See Other",
    "304": "Not Modified",
    "305": "Use Proxy",
    "307": "Temporary Redirect",
    "308": "Permanent Redirect",
    "400": "Bad Request",
    "401": "Unauthorized",
    "402": "Payment Required",
    "403": "Forbidden",
    "404": "Not Found",
    "405": "Method Not Allowed",
    "406": "Not Acceptable",
    "407": "Proxy Authentication Required",
    "408": "Request Timeout",
    "409": "Conflict",
    "410": "Gone",
    "411": "Length Required",
    "412": "Precondition Failed",
    "413": "Payload Too Large",
    "414": "URI Too Long",
    "415": "Unsupported Media Type",
    "416": "Range Not Satisfiable",
    "417": "Expectation Failed",
    "421": "Misdirected Request",
    "422": "Unprocessable Entity",
    "423": "Locked",
    "424": "Failed Dependency",
    "425": "Too Early",
    "426": "Upgrade Required",
    "427": "Unassigned",
    "428": "Precondition Required",
    "429": "Too Many Requests",
    "430": "Unassigned",
    "431": "Request Header Fields Too Large",
    "451": "Unavailable For Legal Reasons",
    "500": "Internal Server Error",
    "501": "Not Implemented",
    "502": "Bad Gateway",
    "503": "Service Unavailable",
    "504": "Gateway Timeout",
    "505": "HTTP Version Not Supported",
    "506": "Variant Also Negotiates",
    "507": "Insufficient Storage",
    "508": "Loop Detected",
    "509": "Unassigned",
    "510": "Not Extended",
    "511": "Network Authentication Required",
}



################################################################################
# Dactyl watch mode handler class
################################################################################
from watchdog.events import PatternMatchingEventHandler

from dactyl.common import *

class UpdaterHandler(PatternMatchingEventHandler):
    """Updates to pattern-matched files means rendering."""
    def __init__(self, builder):
        self.builder = builder
        patterns = ["*template-*.html",
                    "*.md",
                    "*code_samples/*"]
        PatternMatchingEventHandler.__init__(self, patterns)

    def on_any_event(self, event):
        logger.debug("watch: got event!")
        # Set builder to bypass errors, because a file temporarily not existing
        #  should not cause watch mode to fail
        self.builder.config.bypass_errors=True
        self.builder.config.load_pages()
        self.builder.build_all()
        logger.info("done rendering")
        self.builder.copy_static()



################################################################################
# Dactyl config-loading module
################################################################################
from dactyl.common import *
from dactyl.version import __version__
from dactyl.page import DactylPage

# Used to import filters.
from importlib import import_module
import importlib.util

# Not the file containing defaults, but the default name of user-specified conf
DEFAULT_CONFIG_FILE = "dactyl-config.yml"

class DactylConfig:
    def __init__(self, cli_args):
        """Load config from commandline arguments"""
        self.cli_args = cli_args
        self.set_logging()

        # Don't even bother loading the config file if it's just a version query
        if cli_args.version:
            print("Dactyl version %s" % __version__)
            exit(0)

        self.bypass_errors = cli_args.bypass_errors
        if self.bypass_errors:
            yaml.allow_duplicate_keys = True

        # Start with the default config, then overwrite later
        self.config = yaml.load(resource_stream(__name__, "default-config.yml"))
        self.filters = {}
        if cli_args.config:
            self.load_config_from_file(cli_args.config)
        else:
            logger.debug("No config file specified, trying ./dactyl-config.yml")
            self.load_config_from_file(DEFAULT_CONFIG_FILE)
        self.check_consistency()
        self.load_filters()

        self.page_cache = []


    def set_logging(self):
        if self.cli_args.debug:
            logger.setLevel(logging.DEBUG)
        elif not self.cli_args.quiet:
            logger.setLevel(logging.INFO)


    def load_config_from_file(self, config_file):
        logger.debug("loading config file %s..." % config_file)
        try:
            with open(config_file, "r", encoding="utf-8") as f:
                loaded_config = yaml.load(f)
        except FileNotFoundError as e:
            if config_file == DEFAULT_CONFIG_FILE:
                logger.info("Couldn't read a config file; using generic config")
                loaded_config = {}
            else:
                traceback.print_tb(e.__traceback__)
                exit("Fatal: Config file '%s' not found"%config_file)
        except ruamel.yaml.parser.ParserError as e:
            traceback.print_tb(e.__traceback__)
            exit("Fatal: Error parsing config file: %s"%e)

        # Migrate legacy config fields
        if "pdf_template" in loaded_config:
            if "default_pdf_template" in loaded_config:
                recoverable_error("Ignoring redundant global config option "+
                               "pdf_template in favor of default_pdf_template",
                               self.bypass_errors)
            else:
                loaded_config["default_pdf_template"] = loaded_config["pdf_template"]
                logger.warning("Deprecation warning: Global field pdf_template has "
                              +"been renamed default_pdf_template")

        self.config.update(loaded_config)

    def check_consistency(self):
        """
        Check a loaded config for common errors, such as unused pages,
        references to non-existing targets, duplicate target names, etc.
        """

        # Look for poorly or redundantly defined targets
        targetnames = set()
        for t in self.config["targets"]:
            if "name" not in t:
                logger.error("Target does not have required 'name' field: %s" % t)
                exit(1)
            elif t["name"] in targetnames:
                recoverable_error("Duplicate target name in config file: '%s'" %
                    t["name"], self.bypass_errors)
            targetnames.add(t["name"])

        # Check page list for consistency
        for page in self.config["pages"]:
            if "targets" not in page:
                if "name" in page:
                    logger.warning("Page %s is not part of any targets." %
                                 page["name"])
                else:
                    logger.warning("Page %s is not part of any targets." % page)
            elif type(page["targets"]) != list:
                recoverable_error(("targets parameter specified incorrectly; "+
                                  "must be a list. Page: %s") % page,
                                  self.bypass_errors)
            elif set(page["targets"]).difference(targetnames):
                recoverable_error("Page '%s' contains undefined targets: %s" %
                            (page, set(page["targets"]).difference(targetnames)),
                            self.bypass_errors)

    def load_pages(self):
        """
        Preload all config'd pages, not just target pages, to get them
        default name values.
        """

        self.page_cache = []
        skip_pp = self.config.get("skip_preprocessor", False)
        for page_data in self.config["pages"]:
            if OPENAPI_SPEC_KEY not in page_data:
                # Try loading page now; but if it fails, that might be OK
                # depending on whether we need the file later.
                try:
                    self.page_cache.append(DactylPage(self, page_data, skip_pp))
                except Exception as e:
                    logger.warning("Couldn't load page '%s': %s"%(page_data,e))
                    self.page_cache.append(NOT_LOADED_PLACEHOLDER)
            else:
                # OpenAPI specs are too much work to load at this time
                self.page_cache.append(OPENAPI_SPEC_PLACEHOLDER)

    def load_filters(self):
        # Figure out which filters we need
        filternames = set(self.config["default_filters"])
        for target in self.config["targets"]:
            if "filters" in target:
                filternames.update(target["filters"])
        for page in self.config["pages"]:
            if "filters" in page:
                filternames.update(page["filters"])

        # Try loading from custom filter paths in order, fall back to built-ins
        for filter_name in filternames:
            filter_loaded = False
            loading_errors = []
            if "filter_paths" in self.config:
                for filter_path in self.config["filter_paths"]:
                    try:
                        f_filepath = os.path.join(filter_path, "filter_"+filter_name+".py")

                        ## Requires Python 3.5+
                        spec = importlib.util.spec_from_file_location(
                                    "dactyl_filters."+filter_name, f_filepath)
                        self.filters[filter_name] = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(self.filters[filter_name])

                        filter_loaded = True
                        break
                    except FileNotFoundError as e:
                        loading_errors.append({"Path": filter_path, "Error": repr(e)})
                        logger.debug("Filter %s isn't in path %s\nErr:%s" %
                                    (filter_name, filter_path, repr(e)))
                    except Exception as e:
                        loading_errors.append({"Path": filter_path, "Error": repr(e)})
                        recoverable_error("Failed to load filter '%s', with error: %s" %
                                (filter_name, repr(e)), self.bypass_errors)

            if not filter_loaded:
                # Load from the Dactyl module
                try:
                    self.filters[filter_name] = import_module("dactyl.filter_"+filter_name)
                except Exception as e:
                    loading_errors.append({"Path": "(Dactyl Built-ins)", "Error": repr(e)})
                    #logger.debug("Failed to load filter %s. Errors: %s" %
                    #    (filter_name, loading_errors))
                    recoverable_error("Failed to load filter %s. Errors:\n%s" %
                        (filter_name, "\n".join(
                            ["  %s: %s" % (le["Path"], le["Error"])
                                for le in loading_errors])
                        ), self.bypass_errors)

    def load_build_options(self):
        """Overwrites some build-specific options based on the CLI params"""
        if self.cli_args.out_dir:
            self.config["out_path"] = self.cli_args.out_dir

        self.config["skip_preprocessor"] = self.cli_args.skip_preprocessor

        if self.cli_args.template_strict_undefined:
            self.config["template_allow_undefined"] = False
        if self.cli_args.pp_strict_undefined:
            self.config["preprocessor_allow_undefined"] = False


    def __getitem__(self, key):
        return self.config[key]

    def __setitem__(self, key, value):
        self.config[key] = value

    def __contains__(self, key):
        return self.config.__contains__(key)

    def get(self, key, default=None):
        return self.config.get(key, default)
